""" websocket transport """
import time
import struct
import errno
import gevent
from gevent.queue import Empty
from gevent.pywsgi import format_date_time
from hashlib import md5
from socket import SHUT_RDWR, error
from pyramid.response import Response

from pyramid_sockjs import STATE_NEW
from pyramid_sockjs import STATE_OPEN
from pyramid_sockjs import STATE_CLOSING
from pyramid_sockjs import STATE_CLOSED
from pyramid_sockjs.transports import StopStreaming
from pyramid_sockjs.protocol import OPEN, HEARTBEAT
from pyramid_sockjs.protocol import encode, decode, close_frame, message_frame


TIMING = 5.0


def WebSocketTransport(session, request):
    socket = request.environ['gunicorn.socket']
    websocket = request.environ['wsgi.websocket']

    def send():
        if session.state == STATE_NEW:
            try:
                websocket.send('o')
            except:
                return
            session.open()

        if session.state == STATE_CLOSING:
            websocket.send(close_frame(3000, 'Go away!'))
            websocket.close()
            session.closed()
            return

        while True:
            try:
                message = [session.get_transport_message(timeout=TIMING)]
            except Empty:
                message = 'h'
                session.heartbeat()
            else:
                message = message_frame(message)

            if session.state == STATE_CLOSING:
                try:
                    websocket.send(close_frame(3000, 'Go away!'))
                    websocket.close()
                except:
                    pass
                session.closed()
                break

            try:
                websocket.send(message)
            except:
                session.closed()
                break

    def receive():
        while True:
            try:
                message = websocket.receive()
            except:
                session.closed()
                break

            if session.state == STATE_CLOSING:
                try:
                    websocket.send(close_frame(3000, 'Go away!'))
                    websocket.close()
                except:
                    pass
                session.closed()
                break

            if message == '':
                continue

            if message is None:
                session.closed()
                websocket.close()
                break

            try:
                if message.startswith('['):
                    message = message[1:-1]
                decoded_message = decode(message)
            except:
                try:
                    websocket.close(message='broken json')
                except:
                    import traceback
                    traceback.print_exc()
                session.closed()
                break

            if decoded_message:
                session.message(decoded_message)

        session.release()

    return WebSocketResponse(session, request, send, receive, websocket)


def RawWebSocketTransport(session, request):
    socket = request.environ['gunicorn.socket']
    websocket = request.environ['wsgi.websocket']

    def send():
        if session.state == STATE_NEW:
            session.open()

        while True:
            if session.state == STATE_CLOSING:
                try:
                    websocket.close()
                except:
                    pass
                session.closed()
                break

            try:
                message = session.get_transport_message(timeout=TIMING)
            except Empty:
                continue

            if session.state != STATE_OPEN:
                try:
                    websocket.close()
                except:
                    pass
                session.closed()
                break
            else:
                try:
                    websocket.send(message)
                except:
                    session.closed()
                    break

    def receive():
        while True:
            try:
                message = websocket.receive()
            except:
                session.closed()
                break

            if session.state == STATE_CLOSED:
                break

            if session.state == STATE_CLOSING:
                try:
                    websocket.close()
                except:
                    pass
                session.closed()
                break

            if message is None:
                session.closed()
                websocket.close()
                break

            session.message(message)

        session.release()

    return WebSocketResponse(session, request, send, receive, websocket)


class WebSocketResponse(Response):

    def __init__(self, session, request, send, receive, websocket):
        self.__dict__.update(request.response.__dict__)
        self.session = session
        self.request = request
        self.websocket = websocket
        self.send = send
        self.receive = receive

    def __call__(self, environ, start_response):
        # WebsocketHixie76 handshake (test_haproxy)
        if environ['wsgi.websocket_version'] == 'hixie-76':
            part1, part2, socket = environ['wsgi.hixie-keys']

            towrite = [
                'HTTP/1.1 %s\r\n'%self.status,
                'Date: %s\r\n'%format_date_time(time.time())]

            for header in self._abs_headerlist(environ):
                towrite.append("%s: %s\r\n" % header)

            towrite.append("\r\n")
            socket.sendall(''.join(towrite))

            key3 = environ['wsgi.input'].read(8)
            if not key3:
                key3 = environ['wsgi.input'].rfile.read(8)

            socket.sendall(
                md5(struct.pack("!II", part1, part2) + key3).digest())
        else:
            write = start_response(
                self.status, self._abs_headerlist(environ))
            write(self.body)

        try:
            self.session.acquire(self.request)
        except: # should use specific exception
            self.websocket.send('o')
            self.websocket.send(
                close_frame(2010, "Another connection still open", '\n'))
            self.websocket.close()
            return StopStreaming()

        gevent.joinall((gevent.spawn(self.send), gevent.spawn(self.receive)))
        raise StopStreaming()



""" iframe-htmlfile transport """
from gevent.queue import Empty
from pyramid.response import Response

from pyramid_sockjs import STATE_NEW
from pyramid_sockjs import STATE_OPEN
from pyramid_sockjs import STATE_CLOSING
from pyramid_sockjs.transports import StopStreaming
from pyramid_sockjs.protocol import HEARTBEAT
from pyramid_sockjs.protocol import encode, decode, close_frame, message_frame

from .utils import session_cookie


class EventsourceTransport(Response):

    timing = 5.0
    maxsize = 131072 # 128K bytes

    def __init__(self, session, request):
        self.__dict__.update(request.response.__dict__)
        self.session = session
        self.request = request

    def __call__(self, environ, start_response):
        write = start_response(
            self.status,
            (('Content-Type','text/event-stream; charset=UTF-8'),
             ('Cache-Control',
              'no-store, no-cache, must-revalidate, max-age=0'),
             session_cookie(self.request)))
        write('\r\n')

        timing = self.timing
        session = self.session

        if session.state == STATE_NEW:
            write("data: o\r\n\r\n")
            session.open()

        size = 0

        try:
            while True:
                try:
                    message = [session.get_transport_message(timeout=timing)]
                except Empty:
                    message = HEARTBEAT
                    session.heartbeat()
                else:
                    message = message_frame(message)

                if session.state == STATE_CLOSING:
                    write("data: %s\r\n\r\n"%close_frame(1000, 'Go away!'))
                    session.closed()
                    break

                if session.state != STATE_OPEN:
                    break

                try:
                    write("data: %s\r\n\r\n" % message)
                except:
                    session.closed()
                    raise StopStreaming()

                size += len(message)
                if size >= self.maxsize:
                    break
        finally:
            session.release()

        return []



""" interactive shell """
import sys
import code
import gevent
from gevent.queue import Queue
from pprint import pprint
from StringIO import StringIO
from pyramid_sockjs import Session, json


class InteractiveShell(Session):

    def __init__(self, id, *args, **kw):
        super(InteractiveShell, self).__init__(id, *args, **kw)

        self.shell = None
        self.cmd_id = 0
        self.shell_worker = None
        self.commands = Queue()

    def start(self):
        if self.shell_worker is None:
            def worker():
                while 1:
                    command = self.commands.get(True)
                    if command is None:
                        break

                    self.cmd_id = command['id']
                    out = StringIO()
                    orig = sys.stdout
                    sys.stdout = out
                    more = self.shell.runsource(command['source'])
                    sys.stdout = orig

                    self.send({'id': self.cmd_id,
                               'complete': True,
                               'more': more,
                               'out': out.getvalue()})

            self.shell_worker = gevent.Greenlet(worker)

        if not self.shell_worker:
            self.shell_worker.start()

    def write(self, data):
        print data,

    def on_open(self):
        self.shell = code.InteractiveInterpreter(
            {'session': self,
             'registry': self.registry,
             'manager': self.manager})

        self.shell.write = self.write
        self.start()

        self.send({'id': 0,
                   'complete': False,
                   'more': False,
                   'out': "Python %s on %s\n" % (sys.version, sys.platform)})

    def on_message(self, msg):
        self.commands.put_nowait(json.loads(msg))

    def on_closed(self):
        if self.shell_worker:
            self.shell_worker.kill()


if __name__ == '__main__':
    """ interactive shell """
    from pyramid.config import Configurator
    from pyramid_sockjs.paster import gevent_server_runner

    config = Configurator()
    config.include('pyramid_sockjs')

    config.add_sockjs_route('shell', '/++shell++', session=InteractiveShell)

    app = config.make_wsgi_app()
    gevent_server_runner(app, {}, host='127.0.0.1')



import urlparse
import httplib_fork as httplib
from ws4py.client.threadedclient import WebSocketClient
import Queue
import socket
import re


class HttpResponse:
    def __init__(self, method, url,
                 headers={}, body=None, async=False, load=True):
        headers = headers.copy()
        u = urlparse.urlparse(url)
        kwargs = {'timeout': 1.0}
        if u.scheme == 'http':
            conn = httplib.HTTPConnection(u.netloc, **kwargs)
        elif u.scheme == 'https':
            conn = httplib.HTTPSConnection(u.netloc, **kwargs)
        else:
            assert False, "Unsupported scheme " + u.scheme
        assert u.fragment == ''
        path = u.path + ('?' + u.query if u.query else '')
        self.conn = conn
        if not body:
            if method is 'POST':
                # The spec says: "Applications SHOULD use this field
                # to indicate the transfer-length of the message-body,
                # unless this is prohibited by the rules in section
                # 4.4."
                # http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.13
                # While httplib sets it only if there is body.
                headers['Content-Length'] = 0
            conn.request(method, path, headers=headers)
        else:
            if isinstance(body, unicode):
                body = body.encode('utf-8')
            conn.request(method, path, headers=headers, body=body)

        if load:
            if not async:
                self._load()
            else:
                self._async_load()

    def _get_status(self):
        return self.res.status
    status = property(_get_status)

    def __getitem__(self, key):
        return self.headers.get(key.lower())

    def _load(self):
        # That works for Content-Length responses.
        self.res = self.conn.getresponse()
        self.headers = dict( (k.lower(), v) for k, v in self.res.getheaders() )
        self.body = self.res.read()
        self.close()

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None

    def _async_load(self):
        # That works for Transfer-Encoding: Chunked
        self.res = self.conn.getresponse()
        self.headers = dict( (k.lower(), v) for k, v in self.res.getheaders() )

    def read(self):
        data =  self.res.read(10240)
        if data:
            return data
        else:
            self.close()
            return None

def old_POST_async(url, **kwargs):
    return HttpResponse('POST', url, async=True, **kwargs)


class WebSocket8Client(object):
    class ConnectionClosedException(Exception): pass

    def __init__(self, url):
        queue = Queue.Queue()
        self.queue = queue
        class IntWebSocketClient(WebSocketClient):
            def received_message(self, m):
                queue.put(unicode(str(m), 'utf-8'))
            def read_from_connection(self, amount):
                r = super(IntWebSocketClient, self).read_from_connection(amount)
                if not r:
                    queue.put(Ellipsis)
                return r
        self.client = IntWebSocketClient(url)
        self.client.connect()

    def close(self):
        if self.client:
            self.client.running = False
            self.client.close()
            self.client.close_connection()
            self.client._th.join()
            self.client = None

    def send(self, data):
        self.client.send(data)

    def recv(self):
        try:
            r = self.queue.get(timeout=1.0)
            if r is Ellipsis:
                raise self.ConnectionClosedException()
            return r
        except:
            self.close()
            raise

def recvline(s):
    b = []
    c = None
    while c != '\n':
        c = s.recv(1)
        b.append( c )
    return ''.join(b)


class CaseInsensitiveDict(object):
    def __init__(self, *args, **kwargs):
        self.lower = {}
        self.d = dict(*args, **kwargs)
        for k in self.d:
            self[k] = self.d[k]

    def __getitem__(self, key, *args, **kwargs):
        pkey = self.lower.setdefault(key.lower(), key)
        return self.d.__getitem__(pkey, *args, **kwargs)

    def __setitem__(self, key, *args, **kwargs):
        pkey = self.lower.setdefault(key.lower(), key)
        return self.d.__setitem__(pkey, *args, **kwargs)

    def items(self):
        for k in self.lower.values():
            yield (k, self[k])

    def __repr__(self): return repr(self.d)
    def __str__(self): return str(self.d)

    def get(self, key, *args, **kwargs):
        pkey = self.lower.setdefault(key.lower(), key)
        return self.d.get(pkey, *args, **kwargs)

    def __contains__(self, key):
        pkey = self.lower.setdefault(key.lower(), key)
        return pkey in self.d

class Response(object):
    def __repr__(self):
        return '<Response HTTP/%s %s %r %r>' % (
            self.http, self.status, self.description, self.headers)

    def __str__(self): return repr(self)

    def __getitem__(self, key):
        return self.headers.get(key)

    def get(self, key, default):
        return self.headers.get(key, default)


class RawHttpConnection(object):
    def __init__(self, url):
        u = urlparse.urlparse(url)
        self.s = socket.create_connection((u.hostname, u.port), timeout=1)

    def request(self, method, url, headers={}, body=None, timeout=1, http="1.1"):
        headers = CaseInsensitiveDict(headers)
        if method == 'POST':
            body = (body or '').encode('utf-8')
        u = urlparse.urlparse(url)
        headers['Host'] = u.hostname + ':' + str(u.port) if u.port else u.hostname
        if body is not None:
            headers['Content-Length'] = str(len(body))

        rel_url = url[ url.find(u.path): ]

        req = ["%s %s HTTP/%s" % (method, rel_url, http)]
        for k, v in headers.items():
            req.append( "%s: %s" % (k, v) )
        req.append('')
        req.append('')
        self.send('\r\n'.join(req))

        if body:
            self.send(body)

        head = recvline(self.s)
        r = re.match(r'HTTP/(?P<version>\S+) (?P<status>\S+) (?P<description>.*)', head)

        resp = Response()
        resp.http = r.group('version')
        resp.status = int(r.group('status'))
        resp.description = r.group('description').rstrip('\r\n')

        resp.headers = CaseInsensitiveDict()
        while True:
            header = recvline(self.s)
            if header in ['\n', '\r\n']:
                break
            k, _, v = header.partition(':')
            resp.headers[k] = v.lstrip().rstrip('\r\n')

        return resp

    def read(self, size=None):
        if size is None:
            # A single packet by default
            return self.s.recv(999999)
        data = []
        while size > 0:
            c = self.s.recv(size)
            if not c:
                raise Exception('Socket closed!')
            size -= len(c)
            data.append( c )
        return ''.join(data)

    def read_till_eof(self):
        data = []
        while True:
            c = self.s.recv(999999)
            if not c:
                break
            data.append( c )
        return ''.join(data)

    def closed(self):
        # To check if socket is being closed, we need to recv and see
        # if the response is empty. If it is not - we're in trouble -
        # abort.
        t = self.s.settimeout(0.1)
        r = self.s.recv(1) == ''
        if not r:
            raise Exception('Socket not closed!')
        self.s.settimeout(t)
        return r

    def read_chunk(self):
        line = recvline(self.s).rstrip('\r\n')
        bytes = int(line, 16) + 2 # Additional \r\n
        return self.read(bytes)[:-2]

    def send(self, data):
        self.s.sendall(data)

    def close(self):
        self.s.close()


def SynchronousHttpRequest(method, url, **kwargs):
    c = RawHttpConnection(url)
    r = c.request(method, url, **kwargs)
    if r.get('Transfer-Encoding', '').lower() == 'chunked':
        chunks = []
        while True:
            chunk = c.read_chunk()
            if len(chunk) == 0:
                break
            chunks.append( chunk )
        r.body = ''.join(chunks)
    elif r.get('Content-Length', ''):
        cl = int(r['Content-Length'])
        r.body = c.read(cl)
    elif 'close' in [k.strip() for k in r.get('Connection', '').lower().split(',')]:
        r.body = c.read_till_eof()
    else:
        # Whitelist statuses that may not need a response
        if r.status in [101, 304, 204]:
            r.body = ''
        else:
            raise Exception(str(r.status) + ' '+str(r.headers) + " No Transfer-Encoding:chunked nor Content-Length nor Connection:Close!")
    c.close()
    return r

def GET(url, **kwargs):
    return SynchronousHttpRequest('GET', url, **kwargs)

def POST(url, **kwargs):
    return SynchronousHttpRequest('POST', url, **kwargs)

def OPTIONS(url, **kwargs):
    return SynchronousHttpRequest('OPTIONS', url, **kwargs)

def AsynchronousHttpRequest(method, url, **kwargs):
    c = RawHttpConnection(url)
    r = c.request(method, url, **kwargs)
    if r.get('Transfer-Encoding', '').lower() == 'chunked':
        def read():
            return c.read_chunk()
        r.read = read
    elif r.get('Content-Length', ''):
        cl = int(r['Content-Length'])
        def read():
            return c.read(cl)
        r.read = read
    elif ('close' in [k.strip() for k in r.get('Connection', '').lower().split(',')]
          or r.status == 101):
        def read():
            return c.read()
        r.read = read
    else:
        raise Exception(str(r.status) + ' '+str(r.headers) + " No Transfer-Encoding:chunked nor Content-Length nor Connection:Close!")
    def close():
        c.close()
    r.close = close
    return r

def GET_async(url, **kwargs):
    return AsynchronousHttpRequest('GET', url, **kwargs)

def POST_async(url, **kwargs):
    return AsynchronousHttpRequest('POST', url, **kwargs)



import logging
import pyramid_sockjs
from gevent.pywsgi import WSGIHandler
from pyramid_sockjs.transports import StopStreaming

orig_get_environ = WSGIHandler.get_environ
orig_handle_error = WSGIHandler.handle_error


def handle_error(self, type, value, tb): # pragma: no cover
    if issubclass(type, StopStreaming):
        del tb
        return

    return orig_handle_error(self, type, value, tb)


def get_environ(self): # pragma: no cover
    env = orig_get_environ(self)
    env['gunicorn.socket'] = self.socket
    if 'HTTP_CONNECTION' not in env:
        env['HTTP_CONNECTION'] = self.headers.get('Connection','')
    return env


def patch_gevent(): # pragma: no cover
    log = logging.getLogger('pyramid_sockjs')

    if WSGIHandler.handle_error is handle_error:
        # skip patch
        return

    WSGIHandler.get_environ = get_environ
    WSGIHandler.handle_error = handle_error

    log.info('Patching gevent WSGIHandler.get_environ')
    log.info('Patching gevent WSGIHandler.handle_error')


###################
# gunicorn
###################

try:
    import gunicorn.util as util
except ImportError: # pragma: no cover
    pass

def process_headers(self, headers): # pragma: no cover
    for name, value in headers:
        assert isinstance(name, basestring), "%r is not a string" % name
        lname = name.lower().strip()
        if lname == "content-length":
            self.response_length = int(value)
        elif util.is_hoppish(name):
            if lname == "connection":
                # handle websocket
                if value.lower().strip() == "upgrade":
                    self.upgrade = True
            elif lname == "upgrade" and value.lower().strip() == "websocket":
                self.headers.append((name.strip(), str(value).strip()))

            # ignore hopbyhop headers
            continue
        self.headers.append((name.strip(), str(value).strip()))


def patch_gunicorn(): # pragma: no cover
    try:
        from gunicorn.http.wsgi import Response
    except ImportError:
        return

    Response.process_headers = process_headers



import gevent
import gevent.monkey
from gevent.pywsgi import WSGIServer


# For paste.deploy server instantiation (egg:pyramid_sockjs#server)
def gevent_server_runner(wsgi_app, global_conf, **kw):
    gevent.monkey.patch_all()

    def runner():
        host = kw.get('host', '0.0.0.0')
        port = int(kw.get('port', 8080))
        server = WSGIServer((host, port), wsgi_app)
        print('Starting Gevent server on http://%s:%s' % (host, port))
        server.serve_forever()

    jobs = [gevent.spawn(runner)]
    gevent.joinall(jobs)



import gevent
from gevent.queue import Empty
from pyramid.compat import url_unquote
from pyramid.response import Response
from pyramid.httpexceptions import HTTPBadRequest

from pyramid_sockjs import STATE_NEW
from pyramid_sockjs import STATE_OPEN
from pyramid_sockjs import STATE_CLOSING
from pyramid_sockjs import STATE_CLOSED
from pyramid_sockjs.transports import StopStreaming
from pyramid_sockjs.protocol import OPEN, MESSAGE, HEARTBEAT
from pyramid_sockjs.protocol import decode, close_frame, message_frame

from .utils import session_cookie, cors_headers, cache_headers


class XHRStreamingTransport(Response):

    timing = 5.0
    maxsize = 131072 # 128K bytes
    open_seq = 'h' *  2048 + '\n'

    def __init__(self, session, request):
        self.__dict__.update(request.response.__dict__)

        self.session = session
        self.request = request

        meth = request.method
        self.headers = (
            ('Content-Type', 'application/javascript; charset=UTF-8'),
            ("Connection", "close"),
            session_cookie(request),
            ) + cors_headers(request)

        if meth not in ('GET', 'POST', 'OPTIONS'):
            raise Exception("No support for such method: %s"%meth)

    def __call__(self, environ, start_response):
        request = self.request
        if request.method == 'OPTIONS':
            self.status = 204
            self.content_type = 'application/javascript; charset=UTF-8'
            self.headerlist.append(
                ("Access-Control-Allow-Methods", "OPTIONS, POST"))
            self.headerlist.extend(cache_headers(request))
            return super(XHRStreamingTransport, self).__call__(
                environ, start_response)

        write = start_response(
            self.status, self._abs_headerlist(environ))
        write(self.open_seq)

        session = self.session

        #if session.state == STATE_CLOSED:
        #    write('%s\n'%close_frame(1002, "Connection interrupted"))
        #    return ()

        try:
            session.acquire(self.request)
        except: # should use specific exception
            write(close_frame(2010, "Another connection still open", '\n'))
            return ()

        if session.state == STATE_NEW:
            write(OPEN)
            session.open()

        if session.state in (STATE_CLOSING, STATE_CLOSED):
            write(close_frame(3000, 'Go away!', '\n'))
            if session.state == STATE_CLOSING:
                session.closed()
            return ()

        stream_size = 0
        timing = self.timing

        try:
            while True:
                try:
                    message = [session.get_transport_message(timeout=timing)]
                except Empty:
                    message = HEARTBEAT
                    session.heartbeat()
                else:
                    message = message_frame(message, '\n')

                if session.state == STATE_CLOSING:
                    write(close_frame(3000, 'Go away!', '\n'))
                    session.closed()
                    raise StopStreaming()

                if session.state != STATE_OPEN:
                    break

                try:
                    write(message)
                except:
                    session.closed()
                    raise StopStreaming()

                stream_size += len(message)
                if stream_size > self.maxsize:
                    break
        finally:
            session.release()

        return []



from pyramid.testing import DummyRequest
from pyramid.response import Response
from pyramid.httpexceptions import \
     HTTPNotFound, HTTPBadRequest, HTTPMethodNotAllowed

from base import BaseTestCase, SocketMock


class WebSoscketHandshake(BaseTestCase):

    def test_websocket_upgrade(self):
        from pyramid_sockjs.websocket import init_websocket
        request = DummyRequest()

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'Can "Upgrade" only to "WebSocket".')

        request.environ['HTTP_UPGRADE'] = 'ssl'

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'Can "Upgrade" only to "WebSocket".')

    def test_connection_upgrade(self):
        from pyramid_sockjs.websocket import init_websocket
        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'close'})

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, '"Connection" must be "Upgrade".')

    def test_websocket_version(self):
        from pyramid_sockjs.websocket import init_websocket

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'keep-alive, upgrade'})

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'Unsupported WebSocket version.')

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'keep-alive, upgrade',
                     'HTTP_SEC_WEBSOCKET_VERSION': '5'})

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'Unsupported WebSocket version.')

    def test_method_type(self):
        from pyramid_sockjs.websocket import init_websocket

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'keep-alive, upgrade',
                     'HTTP_SEC_WEBSOCKET_VERSION': '8'})
        request.method = 'POST'

        res = init_websocket(request)
        self.assertIsInstance(res, Response)
        self.assertEqual(res.status, '405 Method Not Allowed')

    def test_protocol_type(self):
        from pyramid_sockjs.websocket import init_websocket

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'keep-alive, upgrade',
                     'HTTP_SEC_WEBSOCKET_VERSION': '8',
                     'SERVER_PROTOCOL': 'HTTPS/1.1'})
        request.method = 'GET'

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'Protocol is not HTTP')

    def test_protocol_version(self):
        from pyramid_sockjs.websocket import init_websocket

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'keep-alive, upgrade',
                     'HTTP_SEC_WEBSOCKET_VERSION': '8',
                     'SERVER_PROTOCOL': 'HTTP/1.0'})
        request.method = 'GET'

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'HTTP/1.1 is required')

    def test_websocket_key(self):
        from pyramid_sockjs.websocket import init_websocket

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_CONNECTION': 'keep-alive, upgrade',
                     'HTTP_SEC_WEBSOCKET_VERSION': '8',
                     'SERVER_PROTOCOL': 'HTTP/1.1',
                     'HTTP_SEC_WEBSOCKET_KEY': None,})
        request.method = 'GET'

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, 'HTTP_SEC_WEBSOCKET_KEY is invalid key')

    def test_gevent_wsgi_input(self):
        from pyramid_sockjs.websocket import init_websocket

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_SEC_WEBSOCKET_VERSION': '8',
                     'SERVER_PROTOCOL': 'HTTP/1.1',
                     'HTTP_SEC_WEBSOCKET_KEY': '5Jfbk3Hf5oLcReU416OxpA==',
                     'HTTP_CONNECTION': 'keep-alive, upgrade'})
        request.method = 'GET'

        res = init_websocket(request)
        self.assertIsInstance(res, HTTPBadRequest)
        self.assertEqual(res.detail, "socket object is not available")

    def test_success(self):
        from pyramid_sockjs.websocket import init_websocket

        class WS(object):
            def __init__(self, rfile, environ):
                self.rfile = rfile
                self.environ = environ

        request = DummyRequest(
            environ={'HTTP_UPGRADE': 'websocket',
                     'HTTP_SEC_WEBSOCKET_VERSION': '8',
                     'SERVER_PROTOCOL': 'HTTP/1.1',
                     'HTTP_SEC_WEBSOCKET_KEY': '5Jfbk3Hf5oLcReU416OxpA==',
                     'HTTP_CONNECTION': 'keep-alive, upgrade',
                     'gunicorn.socket': SocketMock()})
        request.method = 'GET'

        from pyramid_sockjs import websocket
        orig = websocket.WebSocketHybi
        websocket.WebSocketHybi = WS
        try:
            ws = init_websocket(request)
        finally:
            websocket.WebSocketHybi = orig

        environ = request.environ
        self.assertIn('wsgi.websocket_version', environ)
        self.assertEqual(environ['wsgi.websocket_version'], 'hybi-8')

        self.assertIn('wsgi.websocket', environ)
        self.assertIsInstance(environ['wsgi.websocket'], WS)

        response = request.response
        self.assertEqual(response.status, '101 Switching Protocols')
        self.assertEqual(response.headers['Upgrade'], 'websocket')
        self.assertEqual(response.headers['Connection'], 'Upgrade')



import gevent
from gevent.queue import Empty
from pyramid.compat import url_unquote
from pyramid.response import Response
from pyramid.httpexceptions import HTTPBadRequest, HTTPServerError

from pyramid_sockjs import STATE_NEW
from pyramid_sockjs import STATE_OPEN
from pyramid_sockjs import STATE_CLOSING
from pyramid_sockjs import STATE_CLOSED

from pyramid_sockjs.protocol import OPEN, MESSAGE, HEARTBEAT
from pyramid_sockjs.protocol import decode, close_frame, message_frame

from .utils import get_messages, session_cookie, cors_headers, cache_headers


class PollingTransport(object):
    """ Long polling derivative transports, used for XHRPolling and JSONPolling.
    """
    timing = 5.0
    heartbeat = False

    def __call__(self, session, request):
        meth = request.method
        session_cookie(request)
        response = request.response

        if meth in ('GET', 'POST'):
            result = self.process(session, request)
            if result is not None:
                return result
        elif meth == 'OPTIONS':
            response.status = 204
            response.content_type = 'application/javascript; charset=UTF-8'
            response.headerlist.append(
                ("Access-Control-Allow-Methods", "OPTIONS, POST"))
            response.headerlist.extend(cors_headers(request))
            response.headerlist.extend(cache_headers(request))
        else:
            raise Exception("No support for such method: " + meth)

        return response


class XHRPollingTransport(PollingTransport):

    method = 'GET'

    def process(self, session, request):
        response = request.response
        response.content_type = 'application/javascript; charset=UTF-8'
        response.headerlist.extend(cors_headers(request))

        def finish(request):
            session.release()

        request.add_finished_callback(finish)

        try:
            session.acquire(request)
        except: # should use specific exception
            response.body = close_frame(
                2010, "Another connection still open", '\n')
            return

        if session.state == STATE_NEW:
            response.body = OPEN
            session.open()
            return

        if session.state in (STATE_CLOSING, STATE_CLOSED):
            response.body = close_frame(3000, 'Go away!', '\n')
            if session.state == STATE_CLOSING:
                session.closed()
            return

        response.body = get_messages(session, self.timing, self.heartbeat)


class XHRSendPollingTransport(PollingTransport):

    method = 'POST'

    def process(self, session, request):
        response = request.response
        response.content_type = 'text/plain; charset=UTF-8'
        response.headerlist.extend(cors_headers(request))

        data = request.body_file.read()
        if not data:
            return HTTPServerError("Payload expected.")

        try:
            messages = decode(data)
        except:
            return HTTPServerError("Broken JSON encoding.")

        for msg in messages:
            session.message(msg)

        response.status = 204



# pyramid_sockjs transports


class StopStreaming(StopIteration):
    """ Connection has been disconnected. """


from .jsonp import JSONPolling
from .websocket import WebSocketTransport
from .htmlfile import HTMLFileTransport
from .eventsource import EventsourceTransport
from .xhrpolling import XHRPollingTransport
from .xhrpolling import XHRSendPollingTransport
from .xhrstreaming import XHRStreamingTransport

from .utils import session_cookie


handlers = {
    'websocket'    : (True, WebSocketTransport),

    'xhr'          : (True, XHRPollingTransport()),
    'xhr_send'     : (False, XHRSendPollingTransport()),
    'xhr_streaming': (True, XHRStreamingTransport),

    'jsonp'        : (True, JSONPolling),
    'jsonp_send'   : (False, JSONPolling),

    'htmlfile'     : (True, HTMLFileTransport),
    'eventsource'  : (True, EventsourceTransport),
}



# pyramid_sockjs

from pyramid_sockjs.session import Session
from pyramid_sockjs.session import SessionManager

from pyramid_sockjs.session import STATE_NEW
from pyramid_sockjs.session import STATE_OPEN
from pyramid_sockjs.session import STATE_CLOSING
from pyramid_sockjs.session import STATE_CLOSED

from pyramid_sockjs.protocol import json
from pyramid_sockjs.route import get_session_manager


def includeme(cfg):
    from pyramid_sockjs.route import add_sockjs_route
    from pyramid_sockjs.route import GetSessionManager

    def get_manager(request, name=''):
        return GetSessionManager(request.registry)

    cfg.add_directive('add_sockjs_route', add_sockjs_route)
    cfg.set_request_property(get_manager, 'get_sockjs_manager', True)

    # patch gevent
    import pyramid_sockjs.monkey
    pyramid_sockjs.monkey.patch_gevent()
    pyramid_sockjs.monkey.patch_gunicorn()



import sys
import gevent
from pyramid_sockjs.session import Session


class EchoSession(Session):

    def on_message(self, message):
        self.send(message)


class CloseSession(Session):

    def on_open(self):
        self.close()


class BroadcastSession(Session):

    def on_message(self, msg):
        self.manager.broadcast(msg)


if __name__ == '__main__':
    """ Simple sockjs tests server """
    from pyramid.config import Configurator
    from pyramid_sockjs.paster import gevent_server_runner
    from pyramid_sockjs.transports import websocket, jsonp
    from pyramid_sockjs.transports.xhrpolling import PollingTransport
    from pyramid_sockjs.transports.eventsource import EventsourceTransport
    from pyramid_sockjs.transports.xhrstreaming import XHRStreamingTransport
    from pyramid_sockjs.transports.htmlfile import HTMLFileTransport

    HTMLFileTransport.maxsize = 4096
    EventsourceTransport.maxsize = 4096
    XHRStreamingTransport.maxsize = 4096
    XHRStreamingTransport.timing = 0.01
    websocket.TIMING = 0.1
    jsonp.timing = 0.1
    PollingTransport.timing = 0.2

    config = Configurator()
    config.include('pyramid_sockjs')

    config.add_sockjs_route('echo', '/echo', session=EchoSession)
    config.add_sockjs_route('wsoff', '/disabled_websocket_echo', session=EchoSession,
                            disable_transports=('websocket',))

    config.add_sockjs_route('close', '/close', session=CloseSession)
    config.add_sockjs_route('broadcast', '/broadcast', session=BroadcastSession)

    config.add_sockjs_route('cookie', '/cookie_needed_echo',
                            session=EchoSession, cookie_needed=True)

    app = config.make_wsgi_app()

    if len(sys.argv) > 1 and (sys.argv[1] == '-g'):
        from gunicorn.app.pasterapp import paste_server
        paste_server(app, port=8081, worker_class='gevent', workers=1)
    else:
        gevent_server_runner(app, {}, **{'host': '127.0.0.1', 'port': '8081'})



#!/usr/bin/env python

# Http quirks
# ===========
#
# During the work on SockJS few interesting aspects of Http were
# identified. Following tests try to trigger that. If the tests end
# with success - you can be more confident that your web server will
# survive clients that are violating some aspects of http.
#
# This tests aren't really a part of SockJS test suite, it's more
# about verification http quirks.
import unittest2 as unittest
import uuid
import urlparse
import httplib_fork as httplib
import os

test_top_url = os.environ.get('SOCKJS_URL', 'http://localhost:8081')
base_url = test_top_url + '/echo'

def POST_empty(url):
    u = urlparse.urlparse(url)
    if u.scheme == 'http':
        conn = httplib.HTTPConnection(u.netloc)
    elif u.scheme == 'https':
        conn = httplib.HTTPSConnection(u.netloc)
    else:
        assert False, "Unsupported scheme " + u.scheme
    path = u.path + ('?' + u.query if u.query else '')
    conn.request('POST', path)
    res = conn.getresponse()
    headers = dict( (k.lower(), v) for k, v in res.getheaders() )
    body = res.read()
    conn.close()
    return res.status, body, headers

class HttpQuirks(unittest.TestCase):
    def test_emptyContentLengthForPost(self):
        # Doing POST without Content-Length shouldn't break the
        # server (it does break misultin)
        trans_url = base_url + '/000/' + str(uuid.uuid4())
        status, body, _ = POST_empty(trans_url + '/xhr')
        self.assertEqual(body, 'o\n')
        self.assertEqual(status, 200)


if __name__ == '__main__':
    unittest.main()



from gevent.queue import Empty
from datetime import datetime, timedelta

from pyramid_sockjs.protocol import HEARTBEAT, message_frame


def get_messages(session, timeout, heartbeat=True):
    messages = []
    try:
        messages.append(session.get_transport_message(timeout=timeout))
        while True:
            try:
                messages.append(session.get_transport_message(block=False))
            except Empty:
                break
    except Empty:
        if heartbeat:
            messages = HEARTBEAT
            session.heartbeat()
        else:
            messages = 'a[]'
    else:
        messages = message_frame(messages, '\n')

    return messages


def cors_headers(request):
    origin = request.environ.get("HTTP_ORIGIN", '*')
    if origin == 'null':
        origin = '*'
    ac_headers = request.environ.get('HTTP_ACCESS_CONTROL_REQUEST_HEADERS')
    if ac_headers is not None:
        return (('access-control-allow-origin', origin),
               ('access-control-allow-credentials', 'true'),
               ('access-control-allow-headers', ac_headers))
    else:
        return (('access-control-allow-origin', origin),
               ('access-control-allow-credentials', 'true'))


def session_cookie(request):
    cookie = request.cookies.get('JSESSIONID')

    if not cookie:
        cookie = 'dummy'

    request.response.set_cookie('JSESSIONID', cookie)
    return ('Set-Cookie', request.response.headers['Set-Cookie'])


td365 = timedelta(days=365)
td365seconds = int((td365.microseconds +
                    (td365.seconds + td365.days*24*3600) * 10**6) / 10**6)

def cache_headers(request):
    d = datetime.now() + td365

    return (
        ('Access-Control-Max-Age', td365seconds),
        ('Cache-Control', 'max-age=%d, public' % td365seconds),
        ('Expires', d.strftime('%a, %d %b %Y %H:%M:%S')),
        )



import sys
from zope.interface import directlyProvides
from pyramid import testing
from pyramid.config import Configurator
from pyramid.interfaces import IRequest

if sys.version_info[:2] == (2, 6): # pragma: no cover
    from unittest2 import TestCase
else:
    from unittest import TestCase


class SocketMock(object):

    def makefile(self):
        """ """

    def sendall(self, msg):
        """ """


class BaseTestCase(TestCase):

    _include = True
    _auto_include = True
    _settings = {}
    _environ = {
        'wsgi.url_scheme':'http',
        'wsgi.version':(1,0),
        'HTTP_HOST': 'example.com',
        'SCRIPT_NAME': '',
        'PATH_INFO': '/',
        'gunicorn.socket': SocketMock()}

    def setUp(self):
        self.init_pyramid()

    def make_request(self, environ=None, request_iface=IRequest, **kwargs):
        if environ is None:
            environ=self._environ
        request = testing.DummyRequest(environ=dict(environ), **kwargs)
        request.request_iface = IRequest
        return request

    def init_pyramid(self):
        self.request = request = self.make_request()
        self.config = testing.setUp(
            request=request,
            settings=self._settings,
            autocommit=self._auto_include)
        self.config.get_routes_mapper()
        self.registry = self.config.registry
        self.request.registry = self.registry

        if self._include:
            self.config.include('pyramid_sockjs')



""" jsonp transport """
import uuid
from gevent.queue import Empty
from pyramid.compat import url_unquote
from pyramid.httpexceptions import HTTPBadRequest, HTTPServerError

from pyramid_sockjs import STATE_NEW
from pyramid_sockjs import STATE_OPEN
from pyramid_sockjs import STATE_CLOSING
from pyramid_sockjs import STATE_CLOSED
from pyramid_sockjs.protocol import OPEN, HEARTBEAT
from pyramid_sockjs.protocol import encode, decode, close_frame, message_frame

from .utils import session_cookie, cors_headers


timing = 5.0

def JSONPolling(session, request):
    meth = request.method
    response = request.response
    response.headers['Content-Type'] = 'application/javascript; charset=UTF-8'
    response.headerlist.append(
        ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'))
    session_cookie(request)

    if session.state == STATE_NEW:
        callback = request.GET.get('c', None)
        if callback is None:
            return HTTPServerError('"callback" parameter required')

        response.text = '%s("o");\r\n' % callback
        session.open()
        session.release()

    elif meth == "GET":
        callback = request.GET.get('c', None)
        if callback is None:
            return HTTPServerError('"callback" parameter required')

        if session.state in (STATE_CLOSING, STATE_CLOSED):
            response.text = "%s(%s);\r\n"%(
                callback, encode(close_frame(3000, 'Go away!')))
            if session.state == STATE_CLOSING:
                session.closed()
        else:
            messages = []
            try:
                messages.append(session.get_transport_message(timeout=timing))
                while True:
                    try:
                        messages.append(
                            session.get_transport_message(block=False))
                    except Empty:
                        break

            except Empty:
                messages = HEARTBEAT
                session.heartbeat()
            else:
                messages = message_frame(messages)

            response.text = "%s(%s);\r\n"%(callback, encode(messages))

        session.release()

    elif meth == "POST":
        data = request.body_file.read()

        ctype = request.headers.get('Content-Type', '').lower()
        if ctype == 'application/x-www-form-urlencoded':
            if not data.startswith('d='):
                return HTTPServerError("Payload expected.")

            data = url_unquote(data[2:])

        if not data:
            return HTTPServerError("Payload expected.")

        try:
            messages = decode(data)
        except:
            return HTTPServerError("Broken JSON encoding.")

        for msg in messages:
            session.message(msg)

        response.status = 200
        response.headers['Content-Type'] = 'text/plain; charset=UTF-8'
        response.body = 'ok'
    else:
        raise Exception("No support for such method: %s"%meth)

    return response



import gevent
from datetime import datetime, timedelta

from base import BaseTestCase


class SessionTestCase(BaseTestCase):

    def setUp(self):
        super(SessionTestCase, self).setUp()

        self.now = datetime.now()
        test_self = self

        class DateTime(object):

            def now(self):
                return test_self.now

        from pyramid_sockjs import session
        session.datetime = DateTime()

    def tearDown(self):
        from pyramid_sockjs import session
        session.datetime = datetime

        super(SessionTestCase, self).tearDown()

    def test_ctor(self):
        from pyramid_sockjs import Session, STATE_NEW
        session = Session('id')

        self.assertEqual(session.id, 'id')
        self.assertEqual(session.expired, False)
        self.assertEqual(session.expires, self.now + timedelta(seconds=10))

        self.assertEqual(session.hits, 0)
        self.assertEqual(session.heartbeats, 0)
        self.assertEqual(session.state, STATE_NEW)

        session = Session('id', timedelta(seconds=15))

        self.assertEqual(session.id, 'id')
        self.assertEqual(session.expired, False)
        self.assertEqual(session.expires, self.now + timedelta(seconds=15))

    def test_str(self):
        from pyramid_sockjs import Session, STATE_OPEN, STATE_CLOSING
        session = Session('test')
        session.hits = 10
        session.heartbeats = 50
        session.state = STATE_OPEN

        self.assertEqual(str(session),
                         "id='test' connected hits=10 heartbeats=50")

        session.state = STATE_CLOSING
        self.assertEqual(str(session),
                         "id='test' disconnected hits=10 heartbeats=50")

        session.send('msg')
        self.assertEqual(
            str(session),
            "id='test' disconnected queue[1] hits=10 heartbeats=50")

    def test_tick(self):
        from pyramid_sockjs import Session
        session = Session('id')

        self.now = self.now + timedelta(hours=1)
        session.tick()
        self.assertEqual(session.expires, self.now + session.timeout)

    def test_tick_different_timeout(self):
        from pyramid_sockjs import Session
        session = Session('id', timedelta(seconds=20))

        self.now = self.now + timedelta(hours=1)
        session.tick()
        self.assertEqual(session.expires, self.now + timedelta(seconds=20))

    def test_tick_custom(self):
        from pyramid_sockjs import Session
        session = Session('id', timedelta(seconds=20))

        self.now = self.now + timedelta(hours=1)
        session.tick(timedelta(seconds=30))
        self.assertEqual(session.expires, self.now + timedelta(seconds=30))

    def test_heartbeat(self):
        from pyramid_sockjs import Session
        session = Session('id')

        self.assertEqual(session.heartbeats, 0)

        session.heartbeat()
        self.assertEqual(session.heartbeats, 1)

        session.heartbeat()
        self.assertEqual(session.heartbeats, 2)

    def test_expire(self):
        from pyramid_sockjs import Session
        session = Session('id')

        session.expired = False

        session.expire()
        self.assertTrue(session.expired)

    def test_send(self):
        from pyramid_sockjs import Session, protocol
        session = Session('id')

        session.send(['message'])
        self.assertEqual(session.queue.get(), ['message'])

    def test_send_string(self):
        from pyramid_sockjs import Session, protocol
        session = Session('id')

        session.send('message')
        self.assertEqual(session.queue.get(), 'message')

    def test_send_tick(self):
        from pyramid_sockjs import Session, protocol
        session = Session('id')

        self.now = self.now + timedelta(hours=1)

        session.send(['message'])
        self.assertEqual(session.expires, self.now + session.timeout)

    def test_get_transport_message(self):
        from pyramid_sockjs import Session
        session = Session('id')

        session.send('message')
        self.assertEqual(session.get_transport_message(), 'message')

        from gevent.queue import Empty

        self.assertRaises(
            Empty, session.get_transport_message, timeout=0.1)

    def test_get_transport_message_tick(self):
        from pyramid_sockjs import Session
        session = Session('id')

        session.send('message')

        self.now = self.now + timedelta(hours=1)

        session.get_transport_message()

        self.assertEqual(session.expires, self.now + session.timeout)

    def test_open(self):
        from pyramid_sockjs import Session, STATE_OPEN
        session = Session('id')
        session.open()
        self.assertEqual(session.state, STATE_OPEN)

    def test_open_on_open(self):
        from pyramid_sockjs import Session

        opened = []

        class TestSession(Session):
            def on_open(self):
                opened.append(True)

        session = TestSession('id')
        session.open()

        self.assertTrue(opened[0])

    def test_open_on_open_exception(self):
        from pyramid_sockjs import Session, STATE_OPEN
        class TestSession(Session):
            def on_open(self):
                raise Exception()

        session = TestSession('id')
        session.open()
        self.assertEqual(session.state, STATE_OPEN)

    def test_message(self):
        from pyramid_sockjs import Session
        session = Session('id')
        session.open()

        self.now = self.now + timedelta(hours=1)

        session.message('message')

        self.assertEqual(session.expires, self.now + session.timeout)

    def test_message_on_message(self):
        from pyramid_sockjs import Session

        messages = []

        class TestSession(Session):
            def on_message(self, msg):
                messages.append(msg)

        session = TestSession('id')
        session.open()
        session.message('message')

        self.assertEqual(messages[0], 'message')

    def test_message_on_message_exception(self):
        from pyramid_sockjs import Session
        class TestSession(Session):
            def on_message(self, msg):
                raise Exception()

        session = TestSession('id')
        session.open()

        err = None
        try:
            session.message('message')
        except Exception as exc: # pragma: no cover
            err = exc

        self.assertIsNone(err)

    def test_close(self):
        from pyramid_sockjs import Session, STATE_CLOSING
        session = Session('id')
        session.open()
        session.close()
        self.assertFalse(session.expired)
        self.assertEqual(session.state, STATE_CLOSING)

    def test_close_event(self):
        from pyramid_sockjs import Session

        closing = []
        class TestSession(Session):
            def on_close(self):
                closing.append(True)

        session = TestSession('id')
        session.open()
        session.close()
        self.assertTrue(closing[0])

    def test_close_on_message_exception(self):
        from pyramid_sockjs import Session
        class TestSession(Session):
            def on_close(self):
                raise Exception()

        session = TestSession('id')
        session.open()

        err = None
        try:
            session.close()
        except Exception as exc: # pragma: no cover
            err = exc

        self.assertIsNone(err)

    def test_closed(self):
        from pyramid_sockjs import Session, STATE_CLOSED
        session = Session('id')
        session.open()
        session.closed()
        self.assertTrue(session.expired)
        self.assertEqual(session.state, STATE_CLOSED)

    def test_closed_event(self):
        from pyramid_sockjs import Session

        closed = []
        class TestSession(Session):
            def on_closed(self):
                closed.append(True)

        session = TestSession('id')
        session.open()
        session.closed()
        self.assertTrue(closed[0])

    def test_closed_on_message_exception(self):
        from pyramid_sockjs import Session, STATE_CLOSED
        class TestSession(Session):
            def on_closed(self):
                raise Exception()

        session = TestSession('id')
        session.open()

        err = None
        try:
            session.closed()
        except Exception as exc: # pragma: no cover
            err = exc

        self.assertIsNone(err)
        self.assertEqual(session.state, STATE_CLOSED)


class GcThreadTestCase(BaseTestCase):

    def setUp(self):
        super(GcThreadTestCase, self).setUp()

        self.gc_executed = False

        def gc(s):
            self.gc_executed = True # pragma: no cover

        from pyramid_sockjs.session import SessionManager

        self.gc_origin = SessionManager._gc
        SessionManager._gc = gc

    def tearDown(self):
        from pyramid_sockjs.session import SessionManager
        SessionManager._gc = self.gc_origin

        super(GcThreadTestCase, self).tearDown()

    def test_gc_thread(self):
        from pyramid_sockjs.session import SessionManager

        sm = SessionManager('sm', self.registry, gc_cycle=0.1)
        sm.start()
        sm.stop()
        #self.assertTrue(self.gc_executed)


class SessionManagerTestCase(BaseTestCase):

    def setUp(self):
        super(SessionManagerTestCase, self).setUp()

        self.now = datetime.now()
        test_self = self

        class DateTime(object):

            def now(self):
                return test_self.now

        from pyramid_sockjs import session
        session.datetime = DateTime()

    def tearDown(self):
        from pyramid_sockjs import session
        session.datetime = datetime

        super(SessionManagerTestCase, self).tearDown()

    def make_one(self):
        from pyramid_sockjs.session import Session, SessionManager
        return Session, SessionManager('sm', self.registry)

    def test_fresh(self):
        Session, sm = self.make_one()

        sm._add(Session('id'))
        sm._gc()

        self.assertIn('id', sm)

    def test_gc_removed(self):
        Session, sm = self.make_one()

        sm._add(Session('id'))
        del sm['id']

        self.assertEqual(len(sm.pool), 1)
        sm._gc()

        self.assertEqual(len(sm.pool), 0)

    def test_gc_expire(self):
        from pyramid_sockjs import STATE_CLOSED
        Session, sm = self.make_one()

        session = Session('id')
        session.open()

        sm._add(session)

        self.now = session.expires + timedelta(seconds=10)

        sm._gc()
        self.assertNotIn('id', sm)
        self.assertTrue(session.expired)
        self.assertEqual(session.state, STATE_CLOSED)

    def test_gc_expire_acquired(self):
        from pyramid_sockjs import STATE_CLOSED
        Session, sm = self.make_one()

        session = Session('id')
        session.open()

        sm._add(session)
        sm.acquired['id'] = session

        self.now = session.expires + timedelta(seconds=10)

        sm._gc()
        self.assertNotIn('id', sm)
        self.assertNotIn('id', sm.acquired)
        self.assertTrue(session.expired)
        self.assertEqual(session.state, STATE_CLOSED)

    def test_gc_one_expire(self):
        Session, sm = self.make_one()

        session1 = Session('id1')
        session1.open()

        session2 = Session('id2')
        session2.open()

        sm._add(session1)
        sm._add(session2)

        self.now = session1.expires + timedelta(seconds=10)

        session2.tick()

        sm._gc()
        self.assertNotIn('id1', sm)
        self.assertIn('id2', sm)

    def test_add(self):
        Session, sm = self.make_one()
        session = Session('id')

        sm._add(session)
        self.assertIn('id', sm)
        self.assertIs(sm['id'], session)
        self.assertIs(sm.pool[0][1], session)
        self.assertIs(session.manager, sm)
        self.assertIs(session.registry, sm.registry)

    def test_add_expired(self):
        Session, sm = self.make_one()

        session = Session('id')
        session.expire()

        self.assertRaises(ValueError, sm._add, session)

    def test_get(self):
        Session, sm = self.make_one()
        session = Session('id')

        sm._add(session)
        self.assertIs(sm.get('id'), session)

    def test_get_unknown(self):
        Session, sm = self.make_one()
        self.assertRaises(KeyError, sm.get, 'id')

    def test_get_unknown_with_default(self):
        Session, sm = self.make_one()
        default = object()

        item = sm.get('id', default=default)
        self.assertIs(item, default)

    def test_get_with_create(self):
        Session, sm = self.make_one()

        session = sm.get('id', True)
        self.assertIn(session.id, sm)
        self.assertIsInstance(session, Session)

    def test_acquire(self):
        Session, sm = self.make_one()

        session = Session('id')
        sm._add(session)

        self.now = self.now + timedelta(hours=1)

        request = object()

        s = sm.acquire(session, request=request)

        self.assertIs(s, session)
        self.assertIs(s.request, request)
        self.assertIn('id', sm.acquired)
        self.assertTrue(sm.acquired['id'])
        self.assertTrue(sm.is_acquired(s))
        self.assertEqual(session.expires, self.now + timedelta(seconds=10))

    def test_acquire_unknown(self):
        Session, sm = self.make_one()
        session = Session('id')

        self.assertRaises(KeyError, sm.acquire, session)

    def test_acquire_locked(self):
        Session, sm = self.make_one()
        session = Session('id')
        sm._add(session)

        s = sm.acquire(session)

        self.assertRaises(KeyError, sm.acquire, session)

    def test_release(self):
        Session, sm = self.make_one()

        session = sm.get('id', True)

        request = object()

        s = sm.acquire(session, request)
        sm.release(s)

        self.assertNotIn('id', sm.acquired)

    def test_active_sessions(self):
        Session, sm = self.make_one()

        s1 = Session('s1')
        s2 = Session('s2')

        sm._add(s1)
        sm._add(s2)

        s2.expire()

        active = list(sm.active_sessions())
        self.assertEqual(len(active), 1)
        self.assertIn(s1, active)

    def test_broadcast(self):
        Session, sm = self.make_one()

        s1 = Session('s1')
        s2 = Session('s2')
        sm._add(s1)
        sm._add(s2)

        sm.broadcast('msg')
        self.assertEqual(s1.get_transport_message(), 'msg')
        self.assertEqual(s2.get_transport_message(), 'msg')

    def test_clear(self):
        from pyramid_sockjs.session import STATE_CLOSED

        Session, sm = self.make_one()

        s1 = Session('s1')
        s1.open()
        s2 = Session('s2')
        s2.open()

        sm._add(s1)
        sm._add(s2)

        sm.clear()

        self.assertFalse(bool(sm))
        self.assertTrue(s1.expired)
        self.assertTrue(s2.expired)
        self.assertEqual(s1.state, STATE_CLOSED)
        self.assertEqual(s2.state, STATE_CLOSED)



""" iframe-htmlfile transport """
import gevent
from gevent.queue import Empty
from pyramid.compat import url_unquote
from pyramid.response import Response
from pyramid.httpexceptions import HTTPBadRequest, HTTPServerError

from pyramid_sockjs import STATE_NEW
from pyramid_sockjs import STATE_OPEN
from pyramid_sockjs import STATE_CLOSING
from pyramid_sockjs.transports import StopStreaming
from pyramid_sockjs.protocol import HEARTBEAT
from pyramid_sockjs.protocol import encode, decode, close_frame, message_frame

from .utils import get_messages, session_cookie, cors_headers


PRELUDE = r"""
<!doctype html>
<html><head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
</head><body><h2>Don't panic!</h2>
  <script>
    document.domain = document.domain;
    var c = parent.%s;
    c.start();
    function p(d) {c.message(d);};
    window.onload = function() {c.stop();};
  </script>
""".strip()


class HTMLFileTransport(Response):

    timing = 5.0
    maxsize = 131072 # 128K bytes

    def __init__(self, session, request):
        response = request.response
        self.__dict__.update(response.__dict__)
        self.session = session
        self.request = request

        self.headers = (
            ('Content-Type', 'text/html; charset=UTF-8'),
            ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'),
            ("Connection", "close"),
            session_cookie(request),
            ) + cors_headers(request)

    def __call__(self, environ, start_response):
        request = self.request
        callback = request.GET.get('c', None)
        if callback is None:
            self.status = 500
            self.body = '"callback" parameter required'
            return super(HTMLFileTransport, self).__call__(
                environ, start_response)

        write = start_response(self.status, self.headerlist)
        prelude = PRELUDE % callback
        prelude += ' ' * 1024
        write(prelude)

        timing = self.timing
        session = self.session

        if session.state == STATE_NEW:
            write('<script>\np("o");\n</script>\r\n')
            session.open()

        size = 0

        try:
            while True:
                try:
                    message = [session.get_transport_message(timeout=timing)]
                except Empty:
                    message = HEARTBEAT
                    session.heartbeat()
                else:
                    message = message_frame(message)

                if session.state == STATE_CLOSING:
                    write("<script>\np(%s);\n</script>\r\n" %
                          encode(close_frame('Go away!')))
                    session.closed()
                    break

                message = "<script>\np(%s);\n</script>\r\n" % encode(message)
                try:
                    write(message)
                except:
                    session.closed()
                    raise StopStreaming()

                size += len(message)
                if size > self.maxsize:
                    break
        finally:
            session.release()

        return []



#!/usr/bin/env python
"""
[**SockJS-protocol**](https://github.com/sockjs/sockjs-protocol) is an
effort to define a protocol between in-browser
[SockJS-client](https://github.com/sockjs/sockjs-client) and its
server-side counterparts, like
[SockJS-node](https://github.com/sockjs/sockjs-client). This should
help others to write alternative server implementations.


This protocol definition is also a runnable test suite, do run it
against your server implementation. Supporting all the tests doesn't
guarantee that SockJS client will work flawlessly, end-to-end tests
using real browsers are always required.
"""
import os
import time
import json
import re
import unittest2 as unittest
from utils import GET, GET_async, POST, POST_async, OPTIONS, old_POST_async
from utils import WebSocket8Client
from utils import RawHttpConnection
import uuid


# Base URL
# ========

"""
The SockJS server provides one or more SockJS services. The services
are usually exposed with a simple url prefixes, like:
`http://localhost:8000/echo` or
`http://localhost:8000/broadcast`. We'll call this kind of url a
`base_url`. There is nothing wrong with base url being more complex,
like `http://localhost:8000/a/b/c/d/echo`. Base url should
never end with a slash.

Base url is the url that needs to be supplied to the SockJS client.

All paths under base url are controlled by SockJS server and are
defined by SockJS protocol.

SockJS protocol can be using either http or https.

To run this tests server pointed by `base_url` needs to support
following services:

 - `echo` - responds with identical data as received
 - `disabled_websocket_echo` - identical to `echo`, but with websockets disabled
 - `cookie_needed_echo` - identical to `echo`, but with JSESSIONID cookies sent
 - `close` - server immediately closes the session

This tests should not be run more often than once in five seconds -
many tests operate on the same (named) sessions and they need to have
enough time to timeout.
"""
test_top_url = os.environ.get('SOCKJS_URL', 'http://localhost:8081')
base_url = test_top_url + '/echo'
close_base_url = test_top_url + '/close'
wsoff_base_url = test_top_url + '/disabled_websocket_echo'
cookie_base_url = test_top_url + '/cookie_needed_echo'


# Static URLs
# ===========

class Test(unittest.TestCase):
    # We are going to test several `404/not found` pages. We don't
    # define a body or a content type.
    def verify404(self, r):
        self.assertEqual(r.status, 404)

    # In some cases `405/method not allowed` is more appropriate.
    def verify405(self, r):
        self.assertEqual(r.status, 405)
        self.assertFalse(r['content-type'])
        self.assertTrue(r['allow'])
        self.assertFalse(r.body)

    # Multiple transport protocols need to support OPTIONS method. All
    # responses to OPTIONS requests must be cacheable and contain
    # appropriate headers.
    def verify_options(self, url, allowed_methods):
        for origin in [None, 'test', 'null']:
            h = {}
            if origin:
                h['Origin'] = origin
            r = OPTIONS(url, headers=h)
            self.assertEqual(r.status, 204)
            self.assertTrue(re.search('public', r['Cache-Control']))
            self.assertTrue(re.search('max-age=[1-9][0-9]{6}', r['Cache-Control']),
                            "max-age must be large, one year (31536000) is best")
            self.assertTrue(r['Expires'])
            self.assertTrue(int(r['access-control-max-age']) > 1000000)
            self.assertEqual(r['Access-Control-Allow-Methods'], allowed_methods)
            self.assertFalse(r.body)
            self.verify_cors(r, origin)

    def verify_no_cookie(self, r):
        self.assertFalse(r['Set-Cookie'])

    # Most of the XHR/Ajax based transports do work CORS if proper
    # headers are set.
    def verify_cors(self, r, origin=None):
        if origin and origin != 'null':
            self.assertEqual(r['access-control-allow-origin'], origin)
        else:
            self.assertEqual(r['access-control-allow-origin'], '*')
        # In order to get cookies (`JSESSIONID` mostly) flying, we
        # need to set `allow-credentials` header to true.
        self.assertEqual(r['access-control-allow-credentials'], 'true')

    # Sometimes, due to transports limitations we need to request
    # private data using GET method. In such case it's very important
    # to disallow any caching.
    def verify_not_cached(self, r, origin=None):
        self.assertEqual(r['Cache-Control'],
                         'no-store, no-cache, must-revalidate, max-age=0')
        self.assertFalse(r['Expires'])
        self.assertFalse(r['Last-Modified'])


# Greeting url: `/`
# ----------------
class BaseUrlGreeting(Test):
    # The most important part of the url scheme, is without doubt, the
    # top url. Make sure the greeting is valid.
    def test_greeting(self):
        for url in [base_url, base_url + '/']:
            r = GET(url)
            self.assertEqual(r.status, 200)
            self.assertEqual(r['content-type'], 'text/plain; charset=UTF-8')
            self.assertEqual(r.body, 'Welcome to SockJS!\n')
            self.verify_no_cookie(r)

    # Other simple requests should return 404.
    def test_notFound(self):
        for suffix in ['/a', '/a.html', '//', '///', '/a/a', '/a/a/', '/a',
                       '/a/']:
            self.verify404(GET(base_url + suffix))


# IFrame page: `/iframe*.html`
# ----------------------------
class IframePage(Test):
    """
    Some transports don't support cross domain communication
    (CORS). In order to support them we need to do a cross-domain
    trick: on remote (server) domain we serve an simple html page,
    that loads back SockJS client javascript and is able to
    communicate with the server within the same domain.
    """
    iframe_body = re.compile('''
^<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script>
    document.domain = document.domain;
    _sockjs_onload = function\(\){SockJS.bootstrap_iframe\(\);};
  </script>
  <script src="(?P<sockjs_url>[^"]*)"></script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>$
'''.strip())

    # SockJS server must provide this html page.
    def test_simpleUrl(self):
        self.verify(base_url + '/iframe.html')

    # To properly utilize caching, the same content must be served
    # for request which try to version the iframe. The server may want
    # to give slightly different answer for every SockJS client
    # revision.
    def test_versionedUrl(self):
        for suffix in ['/iframe-a.html', '/iframe-.html', '/iframe-0.1.2.html',
                       '/iframe-0.1.2abc-dirty.2144.html']:
            self.verify(base_url + suffix)

    # In some circumstances (`devel` set to true) client library
    # wants to skip caching altogether. That is achieved by
    # supplying a random query string.
    def test_queriedUrl(self):
        for suffix in ['/iframe-a.html?t=1234', '/iframe-0.1.2.html?t=123414',
                       '/iframe-0.1.2abc-dirty.2144.html?t=qweqweq123']:
            self.verify(base_url + suffix)

    # Malformed urls must give 404 answer.
    def test_invalidUrl(self):
        for suffix in ['/iframe.htm', '/iframe', '/IFRAME.HTML', '/IFRAME',
                       '/iframe.HTML', '/iframe.xml', '/iframe-/.html']:
            r = GET(base_url + suffix)
            self.verify404(r)

    # The '/iframe.html' page and its variants must give `200/ok` and be
    # served with 'text/html' content type.
    def verify(self, url):
        r = GET(url)
        self.assertEqual(r.status, 200)
        self.assertEqual(r['content-type'], 'text/html; charset=UTF-8')
        # The iframe page must be strongly cacheable, supply
        # Cache-Control, Expires and Etag headers and avoid
        # Last-Modified header.
        self.assertTrue(re.search('public', r['Cache-Control']))
        self.assertTrue(re.search('max-age=[1-9][0-9]{6}', r['Cache-Control']),
                        "max-age must be large, one year (31536000) is best")
        self.assertTrue(r['Expires'])
        self.assertTrue(r['ETag'])
        self.assertFalse(r['last-modified'])

        # Body must be exactly as specified, with the exception of
        # `sockjs_url`, which should be configurable.
        match = self.iframe_body.match(r.body.strip())
        self.assertTrue(match)
        # `Sockjs_url` must be a valid url and should utilize caching.
        sockjs_url = match.group('sockjs_url')
        self.assertTrue(sockjs_url.startswith('/') or
                        sockjs_url.startswith('http'))
        self.verify_no_cookie(r)
        return r


    # The iframe page must be strongly cacheable. ETag headers must
    # not change too often. Server must support 'if-none-match'
    # requests.
    def test_cacheability(self):
        r1 = GET(base_url + '/iframe.html')
        r2 = GET(base_url + '/iframe.html')
        self.assertEqual(r1['etag'], r2['etag'])
        self.assertTrue(r1['etag']) # Let's make sure ETag isn't None.

        r = GET(base_url + '/iframe.html', headers={'If-None-Match': r1['etag']})
        self.assertEqual(r.status, 304)
        self.assertFalse(r['content-type'])
        self.assertFalse(r.body)

# Info test: `/info`
# ------------------
#
# Warning: this is a replacement of `/chunking_test` functionality
# from SockJS 0.1.
class InfoTest(Test):
    # This url is called before the client starts the session. It's
    # used to check server capabilities (websocket support, cookies
    # requiremet) and to get the value of "origin" setting (currently
    # not used).
    #
    # But more importantly, the call to this url is used to measure
    # the roundtrip time between the client and the server. So, please,
    # do respond to this url in a timely fashin.
    def test_basic(self):
        r = GET(base_url + '/info')
        self.assertEqual(r.status, 200)
        self.assertEqual(r['content-type'],
                         'application/json; charset=UTF-8')
        self.verify_no_cookie(r)
        self.verify_not_cached(r)
        self.verify_cors(r)

        data = json.loads(r.body)
        # Are websockets enabled on the server?
        self.assertEqual(data['websocket'], True)
        # Do transports need to support cookies (ie: for load
        # balancing purposes.
        self.assertTrue(data['cookie_needed'] in  [True, False])
        # List of allowed origins. Currently ignored.
        self.assertEqual(data['origins'], ['*:*'])
        # Source of entropy for random number generator.
        self.assertTrue(type(data['entropy']) in [int, long])

    # As browsers don't have a good entropy source, the server must
    # help with tht. Info url must supply a good, unpredictable random
    # number from the range <0; 2^32-1> to feed the browser.
    def test_entropy(self):
        r1 = GET(base_url + '/info')
        data1 = json.loads(r1.body)
        r2 = GET(base_url + '/info')
        data2 = json.loads(r2.body)
        self.assertTrue(type(data1['entropy']) in [int, long])
        self.assertTrue(type(data2['entropy']) in [int, long])
        self.assertNotEqual(data1['entropy'], data2['entropy'])

    # Info url must support CORS.
    def test_options(self):
        self.verify_options(base_url + '/info', 'OPTIONS, GET')

    # SockJS client may be hosted from file:// url. In practice that
    # means the 'Origin' headers sent by the browser will have a value
    # of a string "null". Unfortunately, just echoing back "null"
    # won't work - browser will understand that as a rejection. We
    # must respond with star "*" origin in such case.
    def test_options_null_origin(self):
        url = base_url + '/info'
        r = OPTIONS(url, headers={'Origin': 'null'})
        self.assertEqual(r.status, 204)
        self.assertFalse(r.body)
        self.assertEqual(r['access-control-allow-origin'], '*')

    # The 'disabled_websocket_echo' service should have websockets
    # disabled.
    def test_disabled_websocket(self):
        r = GET(wsoff_base_url + '/info')
        self.assertEqual(r.status, 200)
        data = json.loads(r.body)
        self.assertEqual(data['websocket'], False)


# Session URLs
# ============

# Top session URL: `/<server>/<session>`
# --------------------------------------
#
# The session between the client and the server is always initialized
# by the client. The client chooses `server_id`, which should be a
# three digit number: 000 to 999. It can be supplied by user or
# randomly generated. The main reason for this parameter is to make it
# easier to configure load balancer - and enable sticky sessions based
# on first part of the url.
#
# Second parameter `session_id` must be a random string, unique for
# every session.
#
# It is undefined what happens when two clients share the same
# `session_id`. It is a client responsibility to choose identifier
# with enough entropy.
#
# Neither server nor client API's can expose `session_id` to the
# application. This field must be protected from the app.
class SessionURLs(Test):

    # The server must accept any value in `server` and `session` fields.
    def test_anyValue(self):
        self.verify('/a/a')
        for session_part in ['/_/_', '/1/1', '/abcdefgh_i-j%20/abcdefg_i-j%20']:
            self.verify(session_part)

    # To test session URLs we're going to use `xhr-polling` transport
    # facilitites.
    def verify(self, session_part):
        r = POST(base_url + session_part + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')

    # But not an empty string, anything containing dots or paths with
    # less or more parts.
    def test_invalidPaths(self):
        for suffix in ['//', '/a./a', '/a/a.', '/./.' ,'/', '///']:
            self.verify404(GET(base_url + suffix + '/xhr'))
            self.verify404(POST(base_url + suffix + '/xhr'))

    # A session is identified by only `session_id`. `server_id` is a
    # parameter for load balancer and must be ignored by the server.
    def test_ignoringServerId(self):
        ''' See Protocol.test_simpleSession for explanation. '''
        session_id = str(uuid.uuid4())
        r = POST(base_url + '/000/' + session_id + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')

        payload = '["a"]'
        r = POST(base_url + '/000/' + session_id + '/xhr_send', body=payload)
        self.assertEqual(r.status, 204)
        self.assertFalse(r.body)

        r = POST(base_url + '/999/' + session_id + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'a["a"]\n')

# Protocol and framing
# --------------------
#
# SockJS tries to stay API-compatible with WebSockets, but not on the
# network layer. For technical reasons SockJS must introduce custom
# framing and simple custom protocol.
#
# ### Framing accepted by the client
#
# SockJS client accepts following frames:
#
# * `o` - Open frame. Every time a new session is established, the
#   server must immediately send the open frame. This is required, as
#   some protocols (mostly polling) can't distinguish between a
#   properly established connection and a broken one - we must
#   convince the client that it is indeed a valid url and it can be
#   expecting further messages in the future on that url.
#
# * `h` - Heartbeat frame. Most loadbalancers have arbitrary timeouts
#   on connections. In order to keep connections from breaking, the
#   server must send a heartbeat frame every now and then. The typical
#   delay is 25 seconds and should be configurable.
#
# * `a` - Array of json-encoded messages. For example: `a["message"]`.
#
# * `c` - Close frame. This frame is send to the browser every time
#   the client asks for data on closed connection. This may happen
#   multiple times. Close frame contains a code and a string explaining
#   a reason of closure, like: `c[3000,"Go away!"]`.
#
# ### Framing accepted by the server
#
# SockJS server does not have any framing defined. All incoming data
# is treated as incoming messages, either single json-encoded messages
# or an array of json-encoded messages, depending on transport.
#
# ### Tests
#
# To explain the protocol we'll use `xhr-polling` transport
# facilities.
class Protocol(Test):
    # When server receives a request with unknown `session_id` it must
    # recognize that as request for a new session. When server opens a
    # new sesion it must immediately send an frame containing a letter
    # `o`.
    def test_simpleSession(self):
        trans_url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(trans_url + '/xhr')
        "New line is a frame delimiter specific for xhr-polling"
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')

        # After a session was established the server needs to accept
        # requests for sending messages.
        "Xhr-polling accepts messages as a list of JSON-encoded strings."
        payload = '["a"]'
        r = POST(trans_url + '/xhr_send', body=payload)
        self.assertEqual(r.status, 204)
        self.assertFalse(r.body)

        '''We're using an echo service - we'll receive our message
        back. The message is encoded as an array 'a'.'''
        r = POST(trans_url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'a["a"]\n')

        # Sending messages to not existing sessions is invalid.
        payload = '["a"]'
        r = POST(base_url + '/000/bad_session/xhr_send', body=payload)
        self.verify404(r)

        # The session must time out after 5 seconds of not having a
        # receiving connection. The server must send a heartbeat frame
        # every 25 seconds. The heartbeat frame contains a single `h`
        # character. This delay may be configurable.
        pass
        # The server must not allow two receiving connections to wait
        # on a single session. In such case the server must send a
        # close frame to the new connection.
        r1 = old_POST_async(trans_url + '/xhr', load=False)
        r2 = POST(trans_url + '/xhr')

        self.assertEqual(r2.body, 'c[2010,"Another connection still open"]\n')
        self.assertEqual(r2.status, 200)

        r1.close()

    # The server may terminate the connection, passing error code and
    # message.
    def test_closeSession(self):
        trans_url = close_base_url + '/000/' + str(uuid.uuid4())
        r = POST(trans_url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')

        r = POST(trans_url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'c[3000,"Go away!"]\n')

        # Until the timeout occurs, the server must constantly serve
        # the close message.
        r = POST(trans_url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'c[3000,"Go away!"]\n')


# WebSocket protocols: `/*/*/websocket`
# -------------------------------------
import websocket

# The most important feature of SockJS is to support native WebSocket
# protocol. A decent SockJS server should support at least the
# following variants:
#
#   - hixie-75 (Chrome 4, Safari 5.0.0)
#   - hixie-76/hybi-00 (Chrome 6, Safari 5.0.1)
#   - hybi-07 (Firefox 6)
#   - hybi-10 (Firefox 7, Chrome 14)
#
class WebsocketHttpErrors(Test):
    # Normal requests to websocket should not succeed.
    def test_httpMethod(self):
        r = GET(base_url + '/0/0/websocket')
        self.assertEqual(r.status, 400)
        self.assertTrue('Can "Upgrade" only to "WebSocket".' in r.body)

    # Some proxies and load balancers can rewrite 'Connection' header,
    # in such case we must refuse connection.
    def test_invalidConnectionHeader(self):
        r = GET(base_url + '/0/0/websocket', headers={'Upgrade': 'WebSocket',
                                                      'Connection': 'close'})
        self.assertEqual(r.status, 400)
        self.assertTrue('"Connection" must be "Upgrade".', r.body)

    # WebSocket should only accept GET
    def test_invalidMethod(self):
        for h in [{'Upgrade': 'WebSocket', 'Connection': 'Upgrade'},
                  {}]:
            r = POST(base_url + '/0/0/websocket', headers=h)
            self.verify405(r)


# Support WebSocket Hixie-76 protocol
class WebsocketHixie76(Test):
    def test_transport(self):
        ws_url = 'ws:' + base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = websocket.create_connection(ws_url)
        self.assertEqual(ws.recv(), u'o')
        ws.send(u'["a"]')
        self.assertEqual(ws.recv(), u'a["a"]')
        ws.close()

    def test_close(self):
        ws_url = 'ws:' + close_base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = websocket.create_connection(ws_url)
        self.assertEqual(ws.recv(), u'o')
        self.assertEqual(ws.recv(), u'c[3000,"Go away!"]')

        # The connection should be closed after the close frame.
        with self.assertRaises(websocket.ConnectionClosedException):
            if ws.recv() is None:
                raise websocket.ConnectionClosedException
        ws.close()

    # Empty frames must be ignored by the server side.
    def test_empty_frame(self):
        ws_url = 'ws:' + base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = websocket.create_connection(ws_url)
        self.assertEqual(ws.recv(), u'o')
        # Server must ignore empty messages.
        ws.send(u'')
        ws.send(u'["a"]')
        self.assertEqual(ws.recv(), u'a["a"]')
        ws.close()

    # For WebSockets, as opposed to other transports, it is valid to
    # reuse `session_id`. The lifetime of SockJS WebSocket session is
    # defined by a lifetime of underlying WebSocket connection. It is
    # correct to have two separate sessions sharing the same
    # `session_id` at the same time.
    def test_reuseSessionId(self):
        on_close = lambda(ws): self.assertFalse(True)

        ws_url = 'ws:' + base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws1 = websocket.create_connection(ws_url, on_close=on_close)
        self.assertEqual(ws1.recv(), u'o')

        ws2 = websocket.create_connection(ws_url, on_close=on_close)
        self.assertEqual(ws2.recv(), u'o')

        ws1.send(u'"a"')
        self.assertEqual(ws1.recv(), u'a["a"]')

        ws2.send(u'"b"')
        self.assertEqual(ws2.recv(), u'a["b"]')

        ws1.close()
        ws2.close()

        # It is correct to reuse the same `session_id` after closing a
        # previous connection.
        ws1 = websocket.create_connection(ws_url)
        self.assertEqual(ws1.recv(), u'o')
        ws1.send(u'"a"')
        self.assertEqual(ws1.recv(), u'a["a"]')
        ws1.close()

    # Verify WebSocket headers sanity. Due to HAProxy design the
    # websocket server must support writing response headers *before*
    # receiving -76 nonce. In other words, the websocket code must
    # work like that:
    #
    # * Receive request headers.
    # * Write response headers.
    # * Receive request nonce.
    # * Write response nonce.
    def test_haproxy(self):
        url = base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws_url = 'ws:' + url
        http_url = 'http:' + url
        origin = '/'.join(http_url.split('/')[:3])

        c = RawHttpConnection(http_url)
        r = c.request('GET', http_url, http='1.1', headers={
                'Connection':'Upgrade',
                'Upgrade':'WebSocket',
                'Origin': origin,
                'Sec-WebSocket-Key1': '4 @1  46546xW%0l 1 5',
                'Sec-WebSocket-Key2': '12998 5 Y3 1  .P00'
                })
        # First check response headers
        self.assertEqual(r.status, 101)
        self.assertEqual(r.headers['connection'].lower(), 'upgrade')
        self.assertEqual(r.headers['upgrade'].lower(), 'websocket')
        self.assertEqual(r.headers['sec-websocket-location'], ws_url)
        self.assertEqual(r.headers['sec-websocket-origin'], origin)
        self.assertFalse('Content-Length' in r.headers)
        # Later send token
        c.send('aaaaaaaa')
        self.assertEqual(c.read()[:16],
                         '\xca4\x00\xd8\xa5\x08G\x97,\xd5qZ\xba\xbfC{')

    # When user sends broken data - broken JSON for example, the
    # server must abruptly terminate the ws connection.
    def test_broken_json(self):
        ws_url = 'ws:' + base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = websocket.create_connection(ws_url)
        self.assertEqual(ws.recv(), u'o')
        ws.send(u'["a')
        with self.assertRaises(websocket.ConnectionClosedException):
            if ws.recv() is None:
                raise websocket.ConnectionClosedException
        ws.close()


# The server must support Hybi-10 protocol
class WebsocketHybi10(Test):
    def test_transport(self):
        trans_url = base_url + '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = WebSocket8Client(trans_url)

        self.assertEqual(ws.recv(), 'o')
        # Server must ignore empty messages.
        ws.send(u'')
        ws.send(u'["a"]')
        self.assertEqual(ws.recv(), 'a["a"]')
        ws.close()

    def test_close(self):
        trans_url = close_base_url + '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = WebSocket8Client(trans_url)
        self.assertEqual(ws.recv(), u'o')
        self.assertEqual(ws.recv(), u'c[3000,"Go away!"]')
        with self.assertRaises(ws.ConnectionClosedException):
            ws.recv()
        ws.close()

    # Verify WebSocket headers sanity. Server must support both
    # Hybi-07 and Hybi-10.
    def test_headersSanity(self):
        for version in ['7', '8', '13']:
            url = base_url.split(':',1)[1] + \
                '/000/' + str(uuid.uuid4()) + '/websocket'
            ws_url = 'ws:' + url
            http_url = 'http:' + url
            origin = '/'.join(http_url.split('/')[:3])
            h = {'Upgrade': 'websocket',
                 'Connection': 'Upgrade',
                 'Sec-WebSocket-Version': version,
                 'Sec-WebSocket-Origin': 'http://asd',
                 'Sec-WebSocket-Key': 'x3JJHMbDL1EzLkh9GBhXDw==',
                 }

            r = GET_async(http_url, headers=h)
            self.assertEqual(r.status, 101)
            self.assertEqual(r['sec-websocket-accept'], 'HSmrc0sMlYUkAGmm5OPpG2HaGWk=')
            self.assertEqual(r['connection'].lower(), 'upgrade')
            self.assertEqual(r['upgrade'].lower(), 'websocket')
            self.assertFalse(r['content-length'])
            r.close()

    # When user sends broken data - broken JSON for example, the
    # server must abruptly terminate the ws connection.
    def test_broken_json(self):
        ws_url = 'ws:' + base_url.split(':',1)[1] + \
                 '/000/' + str(uuid.uuid4()) + '/websocket'
        ws = WebSocket8Client(ws_url)
        self.assertEqual(ws.recv(), u'o')
        ws.send(u'["a')
        with self.assertRaises(ws.ConnectionClosedException):
            ws.recv()
        ws.close()

    # As a fun part, Firefox 6.0.2 supports Websockets protocol '7'. But,
    # it doesn't send a normal 'Connection: Upgrade' header. Instead it
    # sends: 'Connection: keep-alive, Upgrade'. Brilliant.
    def test_firefox_602_connection_header(self):
        url = base_url.split(':',1)[1] + \
            '/000/' + str(uuid.uuid4()) + '/websocket'
        ws_url = 'ws:' + url
        http_url = 'http:' + url
        origin = '/'.join(http_url.split('/')[:3])
        h = {'Upgrade': 'websocket',
             'Connection': 'keep-alive, Upgrade',
             'Sec-WebSocket-Version': '7',
             'Sec-WebSocket-Origin': 'http://asd',
             'Sec-WebSocket-Key': 'x3JJHMbDL1EzLkh9GBhXDw==',
             }
        r = GET_async(http_url, headers=h)
        self.assertEqual(r.status, 101)


# XhrPolling: `/*/*/xhr`, `/*/*/xhr_send`
# ---------------------------------------
#
# The server must support xhr-polling.
class XhrPolling(Test):
    # The transport must support CORS requests, and answer correctly
    # to OPTIONS requests.
    def test_options(self):
        for suffix in ['/xhr', '/xhr_send']:
            self.verify_options(base_url + '/abc/abc' + suffix,
                                'OPTIONS, POST')

    # Test the transport itself.
    def test_transport(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')
        self.assertEqual(r['content-type'],
                         'application/javascript; charset=UTF-8')
        self.verify_cors(r)

        # Xhr transports receive json-encoded array of messages.
        r = POST(url + '/xhr_send', body='["x"]')
        self.assertEqual(r.status, 204)
        self.assertFalse(r.body)
        # The content type of `xhr_send` must be set to `text/plain`,
        # even though the response code is `204`. This is due to
        # Firefox/Firebug behaviour - it assumes that the content type
        # is xml and shouts about it.
        self.assertEqual(r['content-type'], 'text/plain; charset=UTF-8')
        self.verify_cors(r)

        r = POST(url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'a["x"]\n')

    # Publishing messages to a non-existing session must result in
    # a 404 error.
    def test_invalid_session(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr_send', body='["x"]')
        self.verify404(r)

    # The server must behave when invalid json data is send or when no
    # json data is sent at all.
    def test_invalid_json(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')

        r = POST(url + '/xhr_send', body='["x')
        self.assertEqual(r.status, 500)
        self.assertTrue("Broken JSON encoding." in r.body)

        r = POST(url + '/xhr_send', body='')
        self.assertEqual(r.status, 500)
        self.assertTrue("Payload expected." in r.body)

        r = POST(url + '/xhr_send', body='["a"]')
        self.assertFalse(r.body)
        self.assertEqual(r.status, 204)

        r = POST(url + '/xhr')
        self.assertEqual(r.body, 'a["a"]\n')
        self.assertEqual(r.status, 200)

    # The server must accept messages send with different content
    # types.
    def test_content_types(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr')
        self.assertEqual(r.body, 'o\n')

        ctypes = ['text/plain', 'T', 'application/json', 'application/xml', '',
                  'application/json; charset=utf-8', 'text/xml; charset=utf-8',
                  'text/xml']
        for ct in ctypes:
            r = POST(url + '/xhr_send', body='["a"]', headers={'Content-Type': ct})
            self.assertEqual(r.status, 204)
            self.assertFalse(r.body)

        r = POST(url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'a[' + (',').join(['"a"']*len(ctypes)) +']\n')

    # When client sends a CORS request with
    # 'Access-Control-Request-Headers' header set, the server must
    # echo back this header as 'Access-Control-Allow-Headers'. This is
    # required in order to get CORS working. Browser will be unhappy
    # otherwise.
    def test_request_headers_cors(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr',
                 headers={'Access-Control-Request-Headers': 'a, b, c'})
        self.assertEqual(r.status, 200)
        self.verify_cors(r)
        self.assertEqual(r['Access-Control-Allow-Headers'], 'a, b, c')

        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr',
                 headers={'Access-Control-Request-Headers': ''})
        self.assertEqual(r.status, 200)
        self.verify_cors(r)
        self.assertFalse(r['Access-Control-Allow-Headers'])

        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr')
        self.assertEqual(r.status, 200)
        self.verify_cors(r)
        self.assertFalse(r['Access-Control-Allow-Headers'])


# XhrStreaming: `/*/*/xhr_streaming`
# ----------------------------------
class XhrStreaming(Test):
    def test_options(self):
        self.verify_options(base_url + '/abc/abc/xhr_streaming',
                            'OPTIONS, POST')

    def test_transport(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST_async(url + '/xhr_streaming')
        self.assertEqual(r.status, 200)
        self.assertEqual(r['Content-Type'],
                         'application/javascript; charset=UTF-8')
        self.verify_cors(r)

        # The transport must first send 2KiB of `h` bytes as prelude.
        self.assertEqual(r.read(), 'h' *  2048 + '\n')

        self.assertEqual(r.read(), 'o\n')

        r1 = POST(url + '/xhr_send', body='["x"]')
        self.assertEqual(r1.status, 204)
        self.assertFalse(r1.body)

        self.assertEqual(r.read(), 'a["x"]\n')
        r.close()

    def test_response_limit(self):
        # Single streaming request will buffer all data until
        # closed. In order to remove (garbage collect) old messages
        # from the browser memory we should close the connection every
        # now and then. By default we should close a streaming request
        # every 128KiB messages was send. The test server should have
        # this limit decreased to 4096B.
        url = base_url + '/000/' + str(uuid.uuid4())
        r = POST_async(url + '/xhr_streaming')
        self.assertEqual(r.status, 200)
        self.assertTrue(r.read()) # prelude
        self.assertEqual(r.read(), 'o\n')

        # Test server should gc streaming session after 4096 bytes
        # were sent (including framing).
        msg = '"' + ('x' * 128) + '"'
        for i in range(31):
            r1 = POST(url + '/xhr_send', body='[' + msg + ']')
            self.assertEqual(r1.status, 204)
            self.assertEqual(r.read(), 'a[' + msg + ']\n')

        # The connection should be closed after enough data was
        # delivered.
        self.assertFalse(r.read())


# EventSource: `/*/*/eventsource`
# -------------------------------
#
# For details of this protocol framing read the spec:
#
# * [http://dev.w3.org/html5/eventsource/](http://dev.w3.org/html5/eventsource/)
#
# Beware leading spaces.
class EventSource(Test):
    def test_transport(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET_async(url + '/eventsource')
        self.assertEqual(r.status, 200)
        self.assertEqual(r['Content-Type'],
                         'text/event-stream; charset=UTF-8')
        # As EventSource is requested using GET we must be very
        # carefull not to allow it being cached.
        self.verify_not_cached(r)

        # The transport must first send a new line prelude, due to a
        # bug in Opera.
        self.assertEqual(r.read(), '\r\n')

        self.assertEqual(r.read(), 'data: o\r\n\r\n')

        r1 = POST(url + '/xhr_send', body='["x"]')
        self.assertFalse(r1.body)
        self.assertEqual(r1.status, 204)

        self.assertEqual(r.read(), 'data: a["x"]\r\n\r\n')

        # This protocol doesn't allow binary data and we need to
        # specially treat leading space, new lines and things like
        # \x00. But, now the protocol json-encodes everything, so
        # there is no way to trigger this case.
        r1 = POST(url + '/xhr_send', body=r'["  \u0000\n\r "]')
        self.assertFalse(r1.body)
        self.assertEqual(r1.status, 204)

        self.assertEqual(r.read(),
                         'data: a["  \\u0000\\n\\r "]\r\n\r\n')

        r.close()

    def test_response_limit(self):
        # Single streaming request should be closed after enough data
        # was delivered (by default 128KiB, but 4KiB for test server).
        # Although EventSource transport is better, and in theory may
        # not need this mechanism, there are some bugs in the browsers
        # that actually prevent the automatic GC. See:
        #  * https://bugs.webkit.org/show_bug.cgi?id=61863
        #  * http://code.google.com/p/chromium/issues/detail?id=68160
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET_async(url + '/eventsource')
        self.assertEqual(r.status, 200)
        self.assertTrue(r.read()) # prelude
        self.assertEqual(r.read(), 'data: o\r\n\r\n')

        # Test server should gc streaming session after 4096 bytes
        # were sent (including framing).
        msg = '"' + ('x' * 4096) + '"'
        r1 = POST(url + '/xhr_send', body='[' + msg + ']')
        self.assertEqual(r1.status, 204)
        self.assertEqual(r.read(), 'data: a[' + msg + ']\r\n\r\n')

        # The connection should be closed after enough data was
        # delivered.
        self.assertFalse(r.read())


# HtmlFile: `/*/*/htmlfile`
# -------------------------
#
# Htmlfile transport is based on research done by Michael Carter. It
# requires a famous `document.domain` trick. Read on:
#
# * [http://stackoverflow.com/questions/1481251/what-does-document-domain-document-domain-do](http://stackoverflow.com/questions/1481251/what-does-document-domain-document-domain-do)
# * [http://cometdaily.com/2007/11/18/ie-activexhtmlfile-transport-part-ii/](http://cometdaily.com/2007/11/18/ie-activexhtmlfile-transport-part-ii/)
#
class HtmlFile(Test):
    head = r'''
<!doctype html>
<html><head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
</head><body><h2>Don't panic!</h2>
  <script>
    document.domain = document.domain;
    var c = parent.%s;
    c.start();
    function p(d) {c.message(d);};
    window.onload = function() {c.stop();};
  </script>
'''.strip()

    def test_transport(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET_async(url + '/htmlfile?c=%63allback')
        self.assertEqual(r.status, 200)
        self.assertEqual(r['Content-Type'],
                         'text/html; charset=UTF-8')
        # As HtmlFile is requested using GET we must be very careful
        # not to allow it being cached.
        self.verify_not_cached(r)

        d = r.read()
        self.assertEqual(d.strip(), self.head % ('callback',))
        self.assertGreater(len(d), 1024)
        self.assertEqual(r.read(),
                         '<script>\np("o");\n</script>\r\n')

        r1 = POST(url + '/xhr_send', body='["x"]')
        self.assertFalse(r1.body)
        self.assertEqual(r1.status, 204)

        self.assertEqual(r.read(),
                         '<script>\np("a[\\"x\\"]");\n</script>\r\n')
        r.close()

    def test_no_callback(self):
        r = GET(base_url + '/a/a/htmlfile')
        self.assertEqual(r.status, 500)
        self.assertTrue('"callback" parameter required' in r.body)

    def test_response_limit(self):
        # Single streaming request should be closed after enough data
        # was delivered (by default 128KiB, but 4KiB for test server).
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET_async(url + '/htmlfile?c=callback')
        self.assertEqual(r.status, 200)
        self.assertTrue(r.read()) # prelude
        self.assertEqual(r.read(),
                         '<script>\np("o");\n</script>\r\n')

        # Test server should gc streaming session after 4096 bytes
        # were sent (including framing).
        msg = ('x' * 4096)
        r1 = POST(url + '/xhr_send', body='["' + msg + '"]')
        self.assertEqual(r1.status, 204)
        self.assertEqual(r.read(),
                         '<script>\np("a[\\"' + msg + '\\"]");\n</script>\r\n')

        # The connection should be closed after enough data was
        # delivered.
        self.assertFalse(r.read())

# JsonpPolling: `/*/*/jsonp`, `/*/*/jsonp_send`
# ---------------------------------------------
class JsonPolling(Test):
    def test_transport(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET(url + '/jsonp?c=%63allback')
        self.assertEqual(r.status, 200)
        self.assertEqual(r['Content-Type'],
                         'application/javascript; charset=UTF-8')
        # As JsonPolling is requested using GET we must be very
        # carefull not to allow it being cached.
        self.verify_not_cached(r)

        self.assertEqual(r.body, 'callback("o");\r\n')

        r = POST(url + '/jsonp_send', body='d=%5B%22x%22%5D',
                 headers={'Content-Type': 'application/x-www-form-urlencoded'})
        # Konqueror does weird things on 204. As a workaround we need
        # to respond with something - let it be the string `ok`.
        self.assertEqual(r.body, 'ok')
        self.assertEqual(r.status, 200)
        self.assertEqual(r['Content-Type'], 'text/plain; charset=UTF-8')

        r = GET(url + '/jsonp?c=%63allback')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'callback("a[\\"x\\"]");\r\n')


    def test_no_callback(self):
        r = GET(base_url + '/a/a/jsonp')
        self.assertEqual(r.status, 500)
        self.assertTrue('"callback" parameter required' in r.body)

    # The server must behave when invalid json data is send or when no
    # json data is sent at all.
    def test_invalid_json(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.body, 'x("o");\r\n')

        r = POST(url + '/jsonp_send', body='d=%5B%22x',
                 headers={'Content-Type': 'application/x-www-form-urlencoded'})
        self.assertEqual(r.status, 500)
        self.assertTrue("Broken JSON encoding." in r.body)

        for data in ['', 'd=', 'p=p']:
            r = POST(url + '/jsonp_send', body=data,
                     headers={'Content-Type': 'application/x-www-form-urlencoded'})
            self.assertEqual(r.status, 500)
            self.assertTrue("Payload expected." in r.body)

        r = POST(url + '/jsonp_send', body='d=%5B%22b%22%5D',
                 headers={'Content-Type': 'application/x-www-form-urlencoded'})
        self.assertEqual(r.body, 'ok')

        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'x("a[\\"b\\"]");\r\n')

    # The server must accept messages sent with different content
    # types.
    def test_content_types(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.body, 'x("o");\r\n')

        r = POST(url + '/jsonp_send', body='d=%5B%22abc%22%5D',
                 headers={'Content-Type': 'application/x-www-form-urlencoded'})
        self.assertEqual(r.body, 'ok')
        r = POST(url + '/jsonp_send', body='["%61bc"]',
                 headers={'Content-Type': 'text/plain'})
        self.assertEqual(r.body, 'ok')

        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'x("a[\\"abc\\",\\"%61bc\\"]");\r\n')

    def test_close(self):
        url = close_base_url + '/000/' + str(uuid.uuid4())
        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.body, 'x("o");\r\n')

        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.body, 'x("c[3000,\\"Go away!\\"]");\r\n')

        r = GET(url + '/jsonp?c=x')
        self.assertEqual(r.body, 'x("c[3000,\\"Go away!\\"]");\r\n')

# JSESSIONID cookie
# -----------------
#
# All transports except WebSockets need sticky session support from
# the load balancer. Some load balancers enable that only when they
# see `JSESSIONID` cookie. User of a sockjs server must be able to
# opt-in for this functionality - and set this cookie for all the
# session urls.
#
class JsessionidCookie(Test):
    # Verify if info has cookie_needed set.
    def test_basic(self):
        r = GET(cookie_base_url + '/info')
        self.assertEqual(r.status, 200)
        self.verify_no_cookie(r)

        data = json.loads(r.body)
        self.assertEqual(data['cookie_needed'], True)

    # Helper to check cookie validity.
    def verify_cookie(self, r):
        self.assertEqual(r['Set-Cookie'].split(';')[0].strip(),
                         'JSESSIONID=dummy')
        self.assertEqual(r['Set-Cookie'].split(';')[1].lower().strip(),
                         'path=/')

    # JSESSIONID cookie must be set by default
    def test_xhr(self):
        # polling url must set cookies
        url = cookie_base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr')
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')
        self.verify_cookie(r)

        # Cookie must be echoed back if it's already set.
        url = cookie_base_url + '/000/' + str(uuid.uuid4())
        r = POST(url + '/xhr', headers={'Cookie': 'JSESSIONID=abcdef'})
        self.assertEqual(r.status, 200)
        self.assertEqual(r.body, 'o\n')
        self.assertEqual(r['Set-Cookie'].split(';')[0].strip(),
                         'JSESSIONID=abcdef')
        self.assertEqual(r['Set-Cookie'].split(';')[1].lower().strip(),
                         'path=/')

    def test_xhr_streaming(self):
        url = cookie_base_url + '/000/' + str(uuid.uuid4())
        r = POST_async(url + '/xhr_streaming')
        self.assertEqual(r.status, 200)
        self.verify_cookie(r)

    def test_eventsource(self):
        url = cookie_base_url + '/000/' + str(uuid.uuid4())
        r = GET_async(url + '/eventsource')
        self.assertEqual(r.status, 200)
        self.verify_cookie(r)

    def test_htmlfile(self):
        url = cookie_base_url + '/000/' + str(uuid.uuid4())
        r = GET_async(url + '/htmlfile?c=%63allback')
        self.assertEqual(r.status, 200)
        self.verify_cookie(r)

    def test_jsonp(self):
        url = cookie_base_url + '/000/' + str(uuid.uuid4())
        r = GET(url + '/jsonp?c=%63allback')
        self.assertEqual(r.status, 200)
        self.verify_cookie(r)

        self.assertEqual(r.body, 'callback("o");\r\n')

        r = POST(url + '/jsonp_send', body='d=%5B%22x%22%5D',
                 headers={'Content-Type': 'application/x-www-form-urlencoded'})
        self.assertEqual(r.body, 'ok')
        self.assertEqual(r.status, 200)
        self.verify_cookie(r)


# Raw WebSocket url: `/websocket`
# -------------------------------
#
# SockJS protocol defines a bit of higher level framing. This is okay
# when the browser using SockJS-client establishes the connection, but
# it's not really appropriate when the connection is being esablished
# from another program. Although SockJS focuses on server-browser
# communication, it should be straightforward to connect to SockJS
# from command line or some any programming language.
#
# In order to make writing command-line clients easier, we define this
# `/websocket` entry point. This entry point is special and doesn't
# use any additional custom framing, no open frame, no
# heartbeats. Only raw WebSocket protocol.
class RawWebsocket(Test):
    def test_transport(self):
        ws = WebSocket8Client(base_url + '/websocket')
        ws.send(u'Hello world!\uffff')
        self.assertEqual(ws.recv(), u'Hello world!\uffff')
        ws.close()

    def test_close(self):
        ws = WebSocket8Client(close_base_url + '/websocket')
        with self.assertRaises(ws.ConnectionClosedException):
            ws.recv()
        ws.close()



# JSON Unicode Encoding
# =====================
#
# SockJS takes the responsibility of encoding Unicode strings for the
# user.  The idea is that SockJS should properly deliver any valid
# string from the browser to the server and back. This is actually
# quite hard, as browsers do some magical character
# translations. Additionally there are some valid characters from
# JavaScript point of view that are not valid Unicode, called
# surrogates (JavaScript uses UCS-2, which is not really Unicode).
#
# Dealing with unicode surrogates (0xD800-0xDFFF) is quite special. If
# possible we should make sure that server does escape decode
# them. This makes sense for SockJS servers that support UCS-2
# (SockJS-node), but can't really work for servers supporting unicode
# properly (Python).
#
# The browser must escape quite a list of chars, this is due to
# browser mangling outgoing chars on transports like XHR.
escapable_by_client = re.compile(u"[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u2000-\u20ff\ufeff\ufff0-\uffff\x00-\x1f\ufffe\uffff\u0300-\u0333\u033d-\u0346\u034a-\u034c\u0350-\u0352\u0357-\u0358\u035c-\u0362\u0374\u037e\u0387\u0591-\u05af\u05c4\u0610-\u0617\u0653-\u0654\u0657-\u065b\u065d-\u065e\u06df-\u06e2\u06eb-\u06ec\u0730\u0732-\u0733\u0735-\u0736\u073a\u073d\u073f-\u0741\u0743\u0745\u0747\u07eb-\u07f1\u0951\u0958-\u095f\u09dc-\u09dd\u09df\u0a33\u0a36\u0a59-\u0a5b\u0a5e\u0b5c-\u0b5d\u0e38-\u0e39\u0f43\u0f4d\u0f52\u0f57\u0f5c\u0f69\u0f72-\u0f76\u0f78\u0f80-\u0f83\u0f93\u0f9d\u0fa2\u0fa7\u0fac\u0fb9\u1939-\u193a\u1a17\u1b6b\u1cda-\u1cdb\u1dc0-\u1dcf\u1dfc\u1dfe\u1f71\u1f73\u1f75\u1f77\u1f79\u1f7b\u1f7d\u1fbb\u1fbe\u1fc9\u1fcb\u1fd3\u1fdb\u1fe3\u1feb\u1fee-\u1fef\u1ff9\u1ffb\u1ffd\u2000-\u2001\u20d0-\u20d1\u20d4-\u20d7\u20e7-\u20e9\u2126\u212a-\u212b\u2329-\u232a\u2adc\u302b-\u302c\uaab2-\uaab3\uf900-\ufa0d\ufa10\ufa12\ufa15-\ufa1e\ufa20\ufa22\ufa25-\ufa26\ufa2a-\ufa2d\ufa30-\ufa6d\ufa70-\ufad9\ufb1d\ufb1f\ufb2a-\ufb36\ufb38-\ufb3c\ufb3e\ufb40-\ufb41\ufb43-\ufb44\ufb46-\ufb4e]")
#
# The server is able to send much more chars verbatim. But, it can't
# send Unicode surrogates over Websockets, also various \u2xxxx chars
# get mangled. Additionally, if the server is capable of handling
# UCS-2 (ie: 16 bit character size), it should be able to deal with
# Unicode surrogates 0xD800-0xDFFF:
# http://en.wikipedia.org/wiki/Mapping_of_Unicode_characters#Surrogates
escapable_by_server = re.compile(u"[\x00-\x1f\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufff0-\uffff]")

client_killer_string_esc = '"' + ''.join([
        r'\u%04x' % (i) for i in range(65536)
            if escapable_by_client.match(unichr(i))]) + '"'
server_killer_string_esc = '"' + ''.join([
        r'\u%04x'% (i) for i in range(255, 65536)
            if escapable_by_server.match(unichr(i))]) + '"'

class JSONEncoding(Test):
    def test_xhr_server_encodes(self):
        # Make sure that server encodes at least all the characters
        # it's supposed to encode.
        trans_url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(trans_url + '/xhr')
        self.assertEqual(r.body, 'o\n')
        self.assertEqual(r.status, 200)

        payload = '["' + json.loads(server_killer_string_esc) + '"]'
        r = POST(trans_url + '/xhr_send', body=payload)
        self.assertEqual(r.status, 204)

        r = POST(trans_url + '/xhr')
        self.assertEqual(r.status, 200)
        # skip framing, quotes and parenthesis
        recv = r.body.strip()[2:-1]

        # Received string is indeed what we send previously, aka - escaped.
        self.assertEqual(recv, server_killer_string_esc)

    def test_xhr_server_decodes(self):
        # Make sure that server decodes the chars we're customly
        # encoding.
        trans_url = base_url + '/000/' + str(uuid.uuid4())
        r = POST(trans_url + '/xhr')
        self.assertEqual(r.body, 'o\n')
        self.assertEqual(r.status, 200)

        payload = '[' + client_killer_string_esc + ']' # Sending escaped
        r = POST(trans_url + '/xhr_send', body=payload)
        self.assertEqual(r.status, 204)

        r = POST(trans_url + '/xhr')
        self.assertEqual(r.status, 200)
        # skip framing, quotes and parenthesis
        recv = r.body.strip()[2:-1]

        # Received string is indeed what we send previously. We don't
        # really need to know what exactly got escaped and what not.
        a = json.loads(recv)
        b = json.loads(client_killer_string_esc)
        self.assertEqual(a, b)


# Handling close
# ==============
#
# Dealing with session closure is quite complicated part of the
# protocol. The exact details here don't matter that much to the
# client side, but it's good to have a common behaviour on the server
# side.
#
# This is less about defining the protocol and more about sanity
# checking implementations.
class HandlingClose(Test):
    # When server is closing session, it should unlink current
    # request. That means, if a new request appears, it should receive
    # an application close message rather than "Another connection
    # still open" message.
    def test_close_frame(self):
        url = close_base_url + '/000/' + str(uuid.uuid4())
        r1 = POST_async(url + '/xhr_streaming')

        r1.read() # prelude
        self.assertEqual(r1.read(), 'o\n')
        self.assertEqual(r1.read(), 'c[3000,"Go away!"]\n')

        r2 = POST_async(url + '/xhr_streaming')
        r2.read() # prelude
        self.assertEqual(r2.read(), 'c[3000,"Go away!"]\n')

        # HTTP streaming requests should be automatically closed after
        # close.
        self.assertFalse(r1.read())
        self.assertFalse(r2.read())

    def test_close_request(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r1 = POST_async(url + '/xhr_streaming')

        r1.read() # prelude
        self.assertEqual(r1.read(), 'o\n')

        r2 = POST_async(url + '/xhr_streaming')
        r2.read() # prelude
        self.assertEqual(r2.read(), 'c[2010,"Another connection still open"]\n')

        # HTTP streaming requests should be automatically closed after
        # getting the close frame.
        self.assertFalse(r2.read())

    # When a polling request is closed by a network error - not by
    # server, the session should be automatically closed. When there
    # is a network error - we're in an undefined state. Some messages
    # may have been lost, there is not much we can do about it.
    def test_abort_xhr_streaming(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r1 = POST_async(url + '/xhr_streaming')
        r1.read() # prelude
        self.assertEqual(r1.read(), 'o\n')

        # Can't do second polling request now.
        r2 = POST_async(url + '/xhr_streaming')
        r2.read() # prelude
        self.assertEqual(r2.read(), 'c[2010,"Another connection still open"]\n')
        self.assertFalse(r2.read())

        r1.close()

        # Polling request now, after we aborted previous one, should
        # trigger a connection closure. Implementations may close
        # the session and forget the state related. Alternatively
        # they may return a 1002 close message.
        r3 = POST_async(url + '/xhr_streaming')
        r3.read() # prelude
        self.assertTrue(r3.read() in ['o\n', 'c[1002,"Connection interrupted"]\n'])
        r3.close()

    # The same for polling transports
    def test_abort_xhr_polling(self):
        url = base_url + '/000/' + str(uuid.uuid4())
        r1 = POST(url + '/xhr')
        self.assertEqual(r1.body, 'o\n')

        r1 = old_POST_async(url + '/xhr', load=False)

        # Can't do second polling request now.
        r2 = POST(url + '/xhr')
        self.assertEqual(r2.body, 'c[2010,"Another connection still open"]\n')

        r1.close()

        # Polling request now, after we aborted previous one, should
        # trigger a connection closure. Implementations may close
        # the session and forget the state related. Alternatively
        # they may return a 1002 close message.
        r3 = POST(url + '/xhr')
        self.assertTrue(r3.body in ['o\n', 'c[1002,"Connection interrupted"]\n'])

# Http 1.0 and 1.1 chunking
# =========================
#
# There seem to be a lot of confusion about http/1.0 and http/1.1
# content-length and transfer-encoding:chunking headers. Although
# following tests don't really test anything sockjs specific, it's
# good to make sure that the server is behaving about this.
#
# It is not the intention of this test to verify all possible urls -
# merely to check the sanity of http server implementation.  It is
# assumed that the implementator is able to apply presented behaviour
# to other urls served by the sockjs server.
class Http10(Test):
    # We're going to test a greeting url. No dynamic content, just the
    # simplest possible response.
    def test_synchronous(self):
        c = RawHttpConnection(base_url)
        # In theory 'connection:Keep-Alive' isn't a valid http/1.0
        # header, but in this header may in practice be issued by a
        # http/1.0 client:
        # http://www.freesoft.org/CIE/RFC/2068/248.htm
        r = c.request('GET', base_url, http='1.0',
                      headers={'Connection':'Keep-Alive'})
        self.assertEqual(r.status, 200)
        # In practice the exact http version on the response doesn't
        # really matter. Many serves always respond 1.1.
        self.assertTrue(r.http in ['1.0', '1.1'])
        # Transfer-encoding is not allowed in http/1.0.
        self.assertFalse(r.headers.get('transfer-encoding'))

        # There are two ways to give valid response. Use
        # Content-Length (and maybe connection:Keep-Alive) or
        # Connection: close.
        if not r.headers.get('content-length'):
            self.assertEqual(r.headers['connection'].lower(), 'close')
            self.assertEqual(c.read(), 'Welcome to SockJS!\n')
            self.assertTrue(c.closed())
        else:
            self.assertEqual(int(r.headers['content-length']), 19)
            self.assertEqual(c.read(19), 'Welcome to SockJS!\n')
            connection = r.headers.get('connection', '').lower()
            if connection in ['close', '']:
                # Connection-close behaviour is default in http 1.0
                self.assertTrue(c.closed())
            else:
                self.assertEqual(connection, 'keep-alive')
                # We should be able to issue another request on the same connection
                r = c.request('GET', base_url, http='1.0',
                              headers={'Connection':'Keep-Alive'})
                self.assertEqual(r.status, 200)

    def test_streaming(self):
        url = close_base_url + '/000/' + str(uuid.uuid4())
        c = RawHttpConnection(url)
        # In theory 'connection:Keep-Alive' isn't a valid http/1.0
        # header, but in this header may in practice be issued by a
        # http/1.0 client:
        # http://www.freesoft.org/CIE/RFC/2068/248.htm
        r = c.request('POST', url + '/xhr_streaming', http='1.0',
                      headers={'Connection':'Keep-Alive'})
        self.assertEqual(r.status, 200)
        # Transfer-encoding is not allowed in http/1.0.
        self.assertFalse(r.headers.get('transfer-encoding'))
        # Content-length is not allowed - we don't know it yet.
        self.assertFalse(r.headers.get('content-length'))

        # `Connection` should be not set or be `close`. On the other
        # hand, if it is set to `Keep-Alive`, it won't really hurt, as
        # we are confident that neither `Content-Length` nor
        # `Transfer-Encoding` are set.

        # This is a the same logic as HandlingClose.test_close_frame
        self.assertEqual(c.read(2048+1)[0], 'h') # prelude
        self.assertEqual(c.read(2), 'o\n')
        self.assertEqual(c.read(19), 'c[3000,"Go away!"]\n')
        self.assertTrue(c.closed())


class Http11(Test):
    def test_synchronous(self):
        c = RawHttpConnection(base_url)
        r = c.request('GET', base_url, http='1.1',
                      headers={'Connection':'Keep-Alive'})
        # Keepalive is default in http 1.1
        self.assertTrue(r.http, '1.1')
        self.assertTrue(r.headers.get('connection', '').lower() in ['keep-alive', ''],
                         "Your server doesn't support connection:Keep-Alive")
        # Server should use 'Content-Length' or 'Transfer-Encoding'
        if r.headers.get('content-length'):
            self.assertEqual(int(r.headers['content-length']), 19)
            self.assertEqual(c.read(19), 'Welcome to SockJS!\n')
            self.assertFalse(r.headers.get('transfer-encoding'))
        else:
            self.assertEqual(r.headers['transfer-encoding'].lower(), 'chunked')
            self.assertEqual(c.read_chunk(), 'Welcome to SockJS!\n')
            self.assertEqual(c.read_chunk(), '')
        # We should be able to issue another request on the same connection
        r = c.request('GET', base_url, http='1.1',
                      headers={'Connection':'Keep-Alive'})
        self.assertEqual(r.status, 200)

    def test_streaming(self):
        url = close_base_url + '/000/' + str(uuid.uuid4())
        c = RawHttpConnection(url)
        r = c.request('POST', url + '/xhr_streaming', http='1.1',
                      headers={'Connection':'Keep-Alive'})
        self.assertEqual(r.status, 200)
        # Transfer-encoding is required in http/1.1.
        self.assertTrue(r.headers['transfer-encoding'].lower(), 'chunked')
        # Content-length is not allowed.
        self.assertFalse(r.headers.get('content-length'))
        # Connection header can be anything, so don't bother verifying it.

        # This is a the same logic as HandlingClose.test_close_frame
        self.assertEqual(c.read_chunk()[0], 'h') # prelude
        self.assertEqual(c.read_chunk(), 'o\n')
        self.assertEqual(c.read_chunk(), 'c[3000,"Go away!"]\n')
        self.assertEqual(c.read_chunk(), '')


# Footnote
# ========

# Make this script runnable.
if __name__ == '__main__':
    unittest.main()



from pyramid.request import Request
from pyramid.events import NewRequest
from pyramid.exceptions import ConfigurationError

from base import BaseTestCase


class PyramidDirectiveTestCase(BaseTestCase):

    _include = False

    def test_directive_add_sockjs_route(self):
        self.config.include('pyramid_sockjs')

        self.assertTrue(hasattr(self.config, 'add_sockjs_route'))

    def test_directive_get_sockjs_manager(self):
        self.config.include('pyramid_sockjs')

        request = Request(self._environ)
        request.registry = self.registry
        self.registry.notify(NewRequest(request))

        self.assertTrue(hasattr(request, 'get_sockjs_manager'))


class AddSockJSRouteTestCase(BaseTestCase):

    def setUp(self):
        super(AddSockJSRouteTestCase, self).setUp()

        self._views = views = {}
        self._views_info = views_info = {}
        self._routes = routes = {}

        def add_route(name, pattern):
            routes[name] = pattern

        def add_view(route_name=None, view=None,
                     permission=None, decorator=None):
            views[route_name] = view
            views_info[route_name] = (permission, decorator)

        self.config.add_route = add_route
        self.config.add_view = add_view

    def test_basic(self):
        name = ''
        self.config.add_sockjs_route()

        self.assertIn('sockjs-%s'%name, self._routes)
        self.assertIn('sockjs-info-%s'%name, self._routes)
        self.assertIn('sockjs-iframe-%s'%name, self._routes)
        self.assertIn('sockjs-iframe-ver-%s'%name, self._routes)

        self.assertIn(self._routes['sockjs-%s'%name],
                      '/__sockjs__/{server}/{session}/{transport}')
        self.assertIn(self._routes['sockjs-info-%s'%name],
                      '/__sockjs__/info')
        self.assertIn(self._routes['sockjs-iframe-%s'%name],
                      '/__sockjs__/iframe.html')
        self.assertIn(self._routes['sockjs-iframe-ver-%s'%name],
                      '/__sockjs__/iframe{version}.html')

        self.assertIn('sockjs-%s'%name, self._views)
        self.assertIn('sockjs-info-%s'%name, self._views)
        self.assertIn('sockjs-iframe-%s'%name, self._views)
        self.assertIn('sockjs-iframe-ver-%s'%name, self._views)

    def test_permission_decorator(self):
        name = ''
        permission = 1
        decorator = 2

        self.config.add_sockjs_route(permission=permission,
                                     decorator=decorator)

        val = (permission, decorator)

        self.assertEqual(val, self._views_info['sockjs-%s'%name])
        self.assertEqual(val, self._views_info['sockjs-info-%s'%name])
        self.assertEqual(val, self._views_info['sockjs-iframe-%s'%name])
        self.assertEqual(val, self._views_info['sockjs-iframe-ver-%s'%name])

    def test_config_error(self):
        self.config.add_sockjs_route('route')
        self.assertRaises(
            ConfigurationError, self.config.add_sockjs_route, 'route')

    def test_custom_name(self):
        name = 'chat-service'
        self.config.add_sockjs_route(name)

        self.assertIn('sockjs-%s'%name, self._routes)
        self.assertIn('sockjs-info-%s'%name, self._routes)
        self.assertIn('sockjs-iframe-%s'%name, self._routes)
        self.assertIn('sockjs-iframe-ver-%s'%name, self._routes)

        self.assertIn('sockjs-%s'%name, self._views)
        self.assertIn('sockjs-info-%s'%name, self._views)
        self.assertIn('sockjs-iframe-%s'%name, self._views)
        self.assertIn('sockjs-iframe-ver-%s'%name, self._views)

    def test_custom_prefix(self):
        name = ''
        prefix = '/__chat__/'

        self.config.add_sockjs_route(name, prefix=prefix)

        self.assertIn(self._routes['sockjs-%s'%name],
                      '/__chat__/{server}/{session}/{transport}')
        self.assertIn(self._routes['sockjs-info-%s'%name],
                      '/__chat__/info')
        self.assertIn(self._routes['sockjs-iframe-%s'%name],
                      '/__chat__/iframe.html')
        self.assertIn(self._routes['sockjs-iframe-ver-%s'%name],
                      '/__chat__/iframe{version}.html')

    def test_session_manager_name(self):
        import pyramid_sockjs

        name = 'example'
        self.config.add_sockjs_route(name)

        self.assertTrue(hasattr(self.registry, '__sockjs_managers__'))
        self.assertIn(name, self.registry.__sockjs_managers__)
        self.assertIsInstance(
            self.registry.__sockjs_managers__[name],
            pyramid_sockjs.SessionManager)

    def test_get_session_manager_default(self):
        import pyramid_sockjs
        self.config.add_sockjs_route()

        request = Request(self._environ)
        request.registry = self.registry
        self.registry.notify(NewRequest(request))

        sm = request.get_sockjs_manager()
        self.assertIs(self.registry.__sockjs_managers__[''], sm)

    def test_get_session_manager_unknown(self):
        request = Request(self._environ)
        request.registry = self.registry
        self.registry.notify(NewRequest(request))

        self.assertRaises(
            KeyError, request.get_sockjs_manager, 'test')

        self.config.add_sockjs_route()
        self.assertRaises(
            KeyError, request.get_sockjs_manager, 'test')

    def test_get_session_manager_name(self):
        import pyramid_sockjs

        name = 'example'
        self.config.add_sockjs_route(name)

        request = Request(self._environ)
        request.registry = self.registry
        self.registry.notify(NewRequest(request))

        sm = request.get_sockjs_manager(name)

        self.assertIs(self.registry.__sockjs_managers__[name], sm)

    def test_get_session_manager_name_internal(self):
        import pyramid_sockjs

        name = 'example'
        self.config.add_sockjs_route(name)

        from pyramid_sockjs.route import get_session_manager

        sm = get_session_manager(name, self.registry)
        self.assertIs(self.registry.__sockjs_managers__[name], sm)

        self.assertRaises(
            KeyError,
            get_session_manager, 'unknown', object())

    def test_custom_session(self):
        import pyramid_sockjs

        class ChatSession(pyramid_sockjs.Session):
            """ """

        self.config.add_sockjs_route(session=ChatSession)
        self.assertIs(
            self.registry.__sockjs_managers__[''].factory, ChatSession)

    def test_custom_session_manager(self):
        import pyramid_sockjs

        class ChatSessionManager(pyramid_sockjs.SessionManager):
            """ """

        sm = ChatSessionManager('chat', self.registry)

        self.config.add_sockjs_route('chat', session_manager=sm)
        self.assertIs(self.registry.__sockjs_managers__['chat'], sm)

    def test_custom_session_manager_different_names(self):
        import pyramid_sockjs

        class ChatSessionManager(pyramid_sockjs.SessionManager):
            """ """

        sm = ChatSessionManager('chat', self.registry)

        self.assertRaises(
            ConfigurationError,
            self.config.add_sockjs_route, 'other-chat', session_manager=sm)


class SessionManagerRouteUrlTestCase(BaseTestCase):

    def test_session_manager_route_url(self):
        import pyramid_sockjs

        name = 'example'
        self.config.add_sockjs_route(name, prefix='/chat-service')
        self.config.commit()

        request = Request(self._environ)
        request.registry = self.registry
        self.registry.notify(NewRequest(request))

        sm = request.get_sockjs_manager(name)
        self.assertEqual(
            sm.route_url(request), 'http://example.com/chat-service/')



import re
import base64
import struct
from hashlib import md5, sha1
from socket import SHUT_RDWR
from pyramid.httpexceptions import HTTPBadRequest, HTTPMethodNotAllowed
from geventwebsocket.handler import reconstruct_url
from geventwebsocket.websocket import WebSocketHybi, WebSocketHixie

KEY = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
SUPPORTED_VERSIONS = ('13', '8', '7')


class WebSocketHybi(WebSocketHybi):

    def _close(self):
        if self.socket is not None:
            self.socket.shutdown(SHUT_RDWR)
            self.socket = None
            self._write = None

            if not self._reading:
                self.fobj.close()

            self.fobj = None


class WebSocketHixie(WebSocketHixie):

    def __init__(self, socket, environ):
        super(WebSocketHixie, self).__init__(socket, environ)

        self.socket = socket

    def close(self, message=''):
        if self.socket is not None:
            self.socket.shutdown(SHUT_RDWR)
            self.socket = None
            self._write = None
            self.fobj.close()
            self.fobj = None


def init_websocket(request):
    environ = request.environ

    if request.method != "GET":
        request.response.status = 405
        request.response.headers = (('Allow','GET'),)
        return request.response

    if 'websocket' not in environ.get('HTTP_UPGRADE', '').lower():
        return HTTPBadRequest('Can "Upgrade" only to "WebSocket".')

    if 'upgrade' not in environ.get('HTTP_CONNECTION', '').lower():
        return HTTPBadRequest('"Connection" must be "Upgrade".')

    version = environ.get("HTTP_SEC_WEBSOCKET_VERSION")
    if not version or version not in SUPPORTED_VERSIONS:
        return HTTPBadRequest('Unsupported WebSocket version.')

    environ['wsgi.websocket_version'] = 'hybi-%s' % version

    # check client handshake for validity
    protocol = environ.get('SERVER_PROTOCOL','')
    if not protocol.startswith("HTTP/"):
        return HTTPBadRequest('Protocol is not HTTP')

    if not (environ.get('GATEWAY_INTERFACE','').endswith('/1.1') or \
              protocol.endswith('/1.1')):
        return HTTPBadRequest('HTTP/1.1 is required')

    key = environ.get("HTTP_SEC_WEBSOCKET_KEY")
    if not key or len(base64.b64decode(key)) != 16:
        return HTTPBadRequest('HTTP_SEC_WEBSOCKET_KEY is invalid key')

    # get gevent.socket see pyramid_sockjs.monkey
    socket = environ.get('gunicorn.socket', None)
    if socket is None:
        return HTTPBadRequest("socket object is not available")

    headers = [
        ("Upgrade", "websocket"),
        ("Connection", "Upgrade"),
        ("Sec-WebSocket-Accept", base64.b64encode(sha1(key + KEY).digest()))]
    request.response.headers = headers
    request.response.status = '101 Switching Protocols'

    environ['wsgi.websocket'] = WebSocketHybi(socket, environ)


def get_key_value(key_value):
    key_number = int(re.sub("\\D", "", key_value))
    spaces = re.subn(" ", "", key_value)[1]

    if key_number % spaces != 0:
        raise Exception(
            "key_number %d is not an intergral multiple of spaces %d",
            key_number, spaces)
    else:
        return key_number / spaces


def init_websocket_hixie(request):
    environ = request.environ

    # get gevent.socket see pyramid_sockjs.monkey
    socket = environ.get('gunicorn.socket', None)
    if socket is None:
        return HTTPBadRequest("socket object is not available")

    websocket = WebSocketHixie(socket, environ)
    environ['wsgi.websocket'] = websocket

    key1 = environ.get('HTTP_SEC_WEBSOCKET_KEY1')
    key2 = environ.get('HTTP_SEC_WEBSOCKET_KEY2')

    if key1 is not None:
        environ['wsgi.websocket_version'] = 'hixie-76'
        if not key1:
            return HTTPBadRequest('SEC-WEBSOCKET-KEY1 header is empty')
        if not key2:
            return HTTPBadRequest('SEC-WEBSOCKET-KEY2 header is empty')

        try:
            part1 = get_key_value(key1)
            part2 = get_key_value(key2)
            environ['wsgi.hixie-keys'] = (part1, part2, socket)
        except Exception as err:
            return HTTPBadRequest(str(err))

        headers = [
            ("Upgrade", "WebSocket"),
            ("Connection", "Upgrade"),
            ("Sec-WebSocket-Location", reconstruct_url(environ)),
            ]
        if websocket.protocol is not None:
            headers.append(("Sec-WebSocket-Protocol", websocket.protocol))

        if websocket.origin:
            headers.append(("Sec-WebSocket-Origin", websocket.origin))

        request.response.headers = headers
        request.response.status = '101 Switching Protocols Handshake'
    else:
        environ['wsgi.websocket_version'] = 'hixie-75'
        headers = [
            ("Upgrade", "WebSocket"),
            ("Connection", "Upgrade"),
            ("WebSocket-Location", reconstruct_url(environ)),
            ]

        if websocket.protocol is not None:
            headers.append(("WebSocket-Protocol", websocket.protocol))
        if websocket.origin:
            headers.append(("WebSocket-Origin", websocket.origin))

        request.response.headers = headers
        request.response.status = '101 Switching Protocols Handshake'



# -*- coding: utf-8 -*-
import sys
from pyramid_sockjs.session import Session


class ChatSession(Session):

    def on_open(self):
        self.manager.broadcast("Someone joined.")

    def on_message(self, message):
        self.manager.broadcast(message)

    def on_close(self):
        self.manager.broadcast("Someone left.")


if __name__ == '__main__':
    """ Simple sockjs chat """
    from pyramid.config import Configurator
    from pyramid_sockjs.paster import gevent_server_runner

    config = Configurator()
    config.include('pyramid_sockjs')

    config.add_sockjs_route(prefix='/__sockjs__', session=ChatSession)

    config.add_route('root', '/')
    config.add_view(route_name='root', renderer='__main__:chat.pt')

    app = config.make_wsgi_app()

    if len(sys.argv) > 1 and (sys.argv[1] == '-g'):
        from gunicorn.app.pasterapp import paste_server
        paste_server(app, port=8080, worker_class='gevent', workers=1)
    else:
        gevent_server_runner(app, {}, host='127.0.0.1')



import json
from base import TestCase


class TestProtocol(TestCase):

    def test_encode(self):
        from pyramid_sockjs import protocol

        self.assertEqual(protocol.encode({}), json.dumps({}))
        self.assertEqual(protocol.encode(['test']), json.dumps(['test']))
        self.assertEqual(protocol.encode('"test"'), json.dumps('"test"'))

    def test_decode(self):
        from pyramid_sockjs import protocol

        self.assertEqual(protocol.decode(json.dumps({})), {})
        self.assertEqual(protocol.decode(json.dumps(['test'])), ['test'])
        self.assertEqual(protocol.decode(json.dumps('"test"')), '"test"')

    def test_close_frame(self):
        from pyramid_sockjs import protocol

        msg = protocol.close_frame(1000, 'Internal error')
        self.assertEqual(msg, 'c[1000,"Internal error"]')

    def test_close_frame_endline(self):
        from pyramid_sockjs import protocol

        msg = protocol.close_frame(1000, 'Internal error', '\n')
        self.assertEqual(msg, 'c[1000,"Internal error"]\n')

    def test_message_frame(self):
        from pyramid_sockjs import protocol

        msg = protocol.message_frame(['msg1', 'msg2'])
        self.assertEqual(msg, 'a%s'%protocol.encode(['msg1', 'msg2']))

    def test_message_frame_endline(self):
        from pyramid_sockjs import protocol

        msg = protocol.message_frame(['msg1', 'msg2'], '\n')
        self.assertEqual(msg, 'a%s\n'%protocol.encode(['msg1', 'msg2']))



from pyramid.httpexceptions import HTTPNotFound, HTTPBadRequest

from pyramid_sockjs import protocol

from base import BaseTestCase, SocketMock


class BaseSockjs(BaseTestCase):

    def make_one(self):
        from pyramid_sockjs.route import SockJSRoute
        from pyramid_sockjs.session import SessionManager

        sm = SessionManager('sm', self.registry)
        return SockJSRoute('sm', sm, 'http:sockjs-cdn', ())


class TestSockJSRoute(BaseSockjs):

    def test_info(self):
        route = self.make_one()

        response = route.info(self.request)
        info = protocol.decode(response.body)

        self.assertTrue(info['websocket'])
        self.assertTrue(info['cookie_needed'])

    def test_info_entropy(self):
        route = self.make_one()

        response = route.info(self.request)
        entropy1 = protocol.decode(response.body)['entropy']

        response = route.info(self.request)
        entropy2 = protocol.decode(response.body)['entropy']

        self.assertFalse(entropy1 == entropy2)

    def test_info_options(self):
        route = self.make_one()

        self.request.method = 'OPTIONS'
        response = route.info(self.request)

        self.assertEqual(response.status, '204 No Content')

        headers = response.headers
        self.assertIn('Access-Control-Max-Age', headers)
        self.assertIn('Cache-Control', headers)
        self.assertIn('Expires', headers)
        self.assertIn('Set-Cookie', headers)
        self.assertIn('access-control-allow-credentials', headers)
        self.assertIn('access-control-allow-origin', headers)

    def test_greeting(self):
        route = self.make_one()

        response = route.greeting(self.request)
        self.assertEqual(response.body, 'Welcome to SockJS!\n')

    def test_iframe(self):
        route = self.make_one()

        response = route.iframe(self.request)
        text = """<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script>
    document.domain = document.domain;
    _sockjs_onload = function(){SockJS.bootstrap_iframe();};
  </script>
  <script src="http:sockjs-cdn"></script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>"""

        self.assertEqual(response.body, text)
        self.assertIn('ETag', response.headers)

    def test_iframe_cache(self):
        route = self.make_one()

        self.request.environ['HTTP_IF_NONE_MATCH'] = 'test'
        response = route.iframe(self.request)

        self.assertEqual(response.status, '304 Not Modified')

    def test_handler_unknown_transport(self):
        route = self.make_one()

        self.request.matchdict = {'transport': 'unknown'}
        res = route.handler(self.request)
        self.assertIsInstance(res, HTTPNotFound)

    def test_handler_emptry_session(self):
        route = self.make_one()
        self.request.matchdict = {'transport': 'websocket', 'session': ''}
        self.assertIsInstance(route.handler(self.request), HTTPNotFound)

    def test_handler_bad_session_id(self):
        route = self.make_one()

        self.request.matchdict = {'transport': 'websocket',
                                  'session': 'test.1', 'server': '000'}
        self.assertIsInstance(route.handler(self.request), HTTPNotFound)

    def test_handler_bad_server_id(self):
        route = self.make_one()

        self.request.matchdict = {'transport': 'websocket',
                                  'session': 'test', 'server': 'test.1'}
        self.assertIsInstance(route.handler(self.request), HTTPNotFound)

    def test_new_session_before_read(self):
        route = self.make_one()

        self.request.matchdict = {
            'transport': 'xhr_send', 'session': 'session', 'server': '000'}
        res = route.handler(self.request)
        self.assertIsInstance(res, HTTPNotFound)

    def test_transport(self):
        route = self.make_one()

        self.request.matchdict = {
            'transport': 'xhr', 'session': 'session', 'server': '000'}
        res = route.handler(self.request)
        self.assertIn('wsgi.sockjs_session', self.request.environ)

        from pyramid_sockjs.session import Session
        self.assertIsInstance(
            self.request.environ['wsgi.sockjs_session'], Session)

    def test_fail_transport(self):
        from pyramid_sockjs.transports import handlers
        def fail(session, request):
            raise Exception('Error')

        handlers['test'] = (True, fail)

        route = self.make_one()

        self.request.matchdict = {
            'transport': 'test', 'session': 'session', 'server': '000'}
        res = route.handler(self.request)
        self.assertIsInstance(res, HTTPBadRequest)

        del handlers['test']

    def test_release_session_for_failed_transport(self):
        from pyramid_sockjs.transports import handlers
        def fail(session, request):
            session.acquire()
            raise Exception('Error')

        handlers['test'] = (True, fail)

        route = self.make_one()

        self.request.matchdict = {
            'transport': 'test', 'session': 'session', 'server': '000'}
        res = route.handler(self.request)

        session = route.session_manager['session']
        self.assertFalse(route.session_manager.is_acquired(session))

        del handlers['test']


class TestWebSocketRoute(BaseSockjs):

    def setUp(self):
        super(TestWebSocketRoute, self).setUp()

        # handler
        from pyramid_sockjs import route
        from pyramid_sockjs.transports import handlers
        self.orig = handlers['websocket']

        self.init = []
        def websocket(session, request):
            self.init.append(True)

        handlers['websocket'] = (True, websocket)

        # init
        self.raise_init = False
        def init_websocket(request):
            if self.raise_init:
                return HTTPNotFound('error')

        self.init_orig = route.init_websocket
        route.init_websocket = init_websocket

    def tearDown(self):
        super(TestWebSocketRoute, self).tearDown()

        from pyramid_sockjs import route
        from pyramid_sockjs.transports import handlers

        handlers['websocket'] = self.orig
        route.init_websocket = self.init_orig

    def test_websocket_init(self):
        route = self.make_one()

        self.request.matchdict = {
            'transport': 'websocket', 'session': 'session', 'server': '000'}

        route.handler(self.request)
        self.assertTrue(self.init[0])

    def test_websocket_init2(self):
        route = self.make_one()

        self.request.environ['HTTP_SEC_WEBSOCKET_VERSION'] = '10'
        self.request.matchdict = {
            'transport': 'websocket', 'session': 'session', 'server': '000'}

        route.handler(self.request)
        self.assertTrue(self.init[0])

    def test_websocket_fail_init(self):
        route = self.make_one()

        self.request.matchdict = {
            'transport': 'websocket', 'session': 'session', 'server': '000'}

        self.raise_init = True
        res = route.handler(self.request)
        self.assertIsInstance(res, HTTPNotFound)

    def test_websocket_old_versions(self):
        route = self.make_one()

        self.request.environ['HTTP_ORIGIN'] = 'origin'
        self.request.matchdict = {
            'transport': 'websocket', 'session': 'session', 'server': '000'}

        res = route.handler(self.request)
        self.assertIsNone(res)

    def test_raw_websocket_old_versions(self):
        route = self.make_one()

        self.request.environ['HTTP_ORIGIN'] = 'origin'

        res = route.websocket(self.request)
        self.assertIsInstance(res, HTTPNotFound)

    def test_raw_websocket(self):
        from pyramid_sockjs.transports.websocket import WebSocketResponse

        route = self.make_one()

        self.request.environ['wsgi.websocket'] = object()
        res = route.websocket(self.request)
        self.assertIsInstance(res, WebSocketResponse)

    def test_raw_websocket_fail(self):
        route = self.make_one()

        self.raise_init = True
        res = route.websocket(self.request)
        self.assertIsInstance(res, HTTPNotFound)



import time
import hashlib
from email import utils
from datetime import datetime
from pyramid.compat import string_types

_days = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
_months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
           'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

# json
# -----------

# Fastest
try:
    import ujson as json
    kwargs = {} # pragma: no cover
except ImportError: # pragma: no cover
    def dthandler(obj):
        if isinstance(obj, datetime):
            now = obj.timetuple()
            return '%s, %02d %s %04d %02d:%02d:%02d -0000' % (
                _days[now[6]], now[2],
                _months[now[1] - 1], now[0], now[3], now[4], now[5])

    kwargs = {'default': dthandler, 'separators': (',', ':')}

    # Faster
    try:
        import simplejson as json
    except ImportError:
        # Slowest
        import json


# Frames
# ------

OPEN      = "o\n"
CLOSE     = "c"
MESSAGE   = "a"
HEARTBEAT = "h\n"


# ------------------

IFRAME_HTML = """<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script>
    document.domain = document.domain;
    _sockjs_onload = function(){SockJS.bootstrap_iframe();};
  </script>
  <script src="%s"></script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>
""".strip()

IFRAME_MD5 = hashlib.md5(IFRAME_HTML).hexdigest()

decode = json.loads

def encode(data):
    return json.dumps(data, **kwargs)

def close_frame(code, reason, eol=''):
    return '%s[%d,%s]%s' % (CLOSE, code, encode(reason), eol)

def message_frame(data, eol=''):
    return '%s%s%s'%(MESSAGE, encode(data), eol)



from pyramid_sockjs import Session


class EchoSession(Session):

    def on_message(self, msg):
        self.send(msg)

    @classmethod
    def dump_stats(self):
        print 'Clients: %d' % (len(self.clients))


if __name__ == '__main__':
    from pyramid.config import Configurator
    from pyramid_sockjs.paster import gevent_server_runner

    config = Configurator()
    config.include('pyramid_sockjs')

    config.add_sockjs_route(prefix='/echo', session=EchoSession)

    app = config.make_wsgi_app()
    gevent_server_runner(app, {})
